from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseAction, BaseBatchAction


@dataclass
class ContainerRegistryAction(BaseAction):
    @override
    @classmethod
    def entity_type(cls) -> str:
        return "container_registry"


@dataclass
class ContainerRegistryBatchAction(BaseBatchAction):
    @override
    @classmethod
    def entity_type(cls) -> str:
        return "container_registry"



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Callable

from ai.backend.common.data.notification import NotifiableMessage
from ai.backend.common.events.event_types.notification import NotificationTriggeredEvent
from ai.backend.logging import BraceStyleAdapter

from ...data.notification import NotificationRuleType
from ...services.notification.actions import ProcessNotificationAction

if TYPE_CHECKING:
    from ...services.processors import Processors

log = BraceStyleAdapter(logging.getLogger(__spec__.name))  # type: ignore[name-defined]

__all__ = ("NotificationEventHandler",)


class NotificationEventHandler:
    """
    Event handler for notification events.

    Receives NotificationTriggeredEvent anycast events and delegates
    processing to NotificationProcessors.
    """

    _processors_factory: Callable[[], Processors]

    def __init__(
        self,
        processors_factory: Callable[[], Processors],
    ) -> None:
        self._processors_factory = processors_factory

    async def handle_notification_triggered(
        self,
        context: None,
        source: str,
        event: NotificationTriggeredEvent,
    ) -> None:
        """
        Handle NotificationTriggeredEvent by delegating to the processor.

        Args:
            context: Event context (unused for notifications)
            source: Event source identifier
            event: NotificationTriggeredEvent containing notification data
        """
        log.info(
            "Received notification event: {0} from {1}",
            event.rule_type,
            source,
        )

        # Validate notification_data against the rule type's schema
        rule_type = NotificationRuleType(event.rule_type)
        try:
            validated_data = NotifiableMessage.validate_notification_data(
                rule_type=rule_type,
                data=event.notification_data,
            )
        except Exception as e:
            log.error(
                "Failed to validate notification data for rule type {0}: {1}",
                event.rule_type,
                str(e),
            )
            # Re-raise to let the caller know validation failed
            raise

        # Delegate to processor for business logic
        processors = self._processors_factory()
        await processors.notification.process_notification.wait_for_complete(
            ProcessNotificationAction(
                rule_type=rule_type,
                timestamp=event.timestamp,
                notification_data=validated_data,
            )
        )



"""Repository pattern implementation for schedule operations."""

from __future__ import annotations

import logging
from typing import Mapping, Optional
from uuid import UUID

from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.common.resource.types import TotalResourceData
from ai.backend.common.types import (
    AccessKey,
    AgentId,
    SessionId,
    SlotName,
    SlotTypes,
    VFolderMount,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.image.types import ImageIdentifier
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.exceptions import ErrorStatusInfo
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.sokovan.scheduler.results import ScheduledSessionData
from ai.backend.manager.sokovan.scheduler.types import (
    AllocationBatch,
    KernelCreationInfo,
    SessionRunningData,
    SessionsForPullWithImages,
    SessionsForStartWithImages,
    SessionTransitionData,
)

from .cache_source.cache_source import ScheduleCacheSource
from .db_source.db_source import ScheduleDBSource
from .types.base import SchedulingSpec
from .types.scheduling import SchedulingData
from .types.session import (
    KernelTerminationResult,
    MarkTerminatingResult,
    SessionTerminationResult,
    SweptSessionInfo,
    TerminatingKernelWithAgentData,
    TerminatingSessionData,
)
from .types.session_creation import (
    AllowedScalingGroup,
    SessionCreationContext,
    SessionCreationSpec,
    SessionEnqueueData,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

scheduler_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.SCHEDULER_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=3,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class SchedulerRepository:
    """
    Repository that orchestrates between DB and cache sources for scheduling operations.
    """

    _db_source: ScheduleDBSource
    _cache_source: ScheduleCacheSource
    _config_provider: ManagerConfigProvider

    def __init__(
        self,
        db: ExtendedAsyncSAEngine,
        valkey_stat: ValkeyStatClient,
        config_provider: ManagerConfigProvider,
    ):
        self._db_source = ScheduleDBSource(db)
        self._cache_source = ScheduleCacheSource(valkey_stat)
        self._config_provider = config_provider

    @scheduler_repository_resilience.apply()
    async def get_scheduling_data(self, scaling_group: str) -> Optional[SchedulingData]:
        """
        Get scheduling data from database.
        Returns None if no pending sessions exist.
        Raises ScalingGroupNotFound if scaling group doesn't exist.
        """
        known_slot_types = await self._get_known_slot_types()
        max_container_count = await self._get_max_container_count()
        spec = SchedulingSpec(
            known_slot_types=known_slot_types,
            max_container_count=max_container_count,
        )

        scheduling_data = await self._db_source.get_scheduling_data(scaling_group, spec)
        if not scheduling_data.pending_sessions.sessions:
            return None

        return scheduling_data

    @scheduler_repository_resilience.apply()
    async def allocate_sessions(
        self, allocation_batch: AllocationBatch
    ) -> list[ScheduledSessionData]:
        """
        Allocate sessions by updating DB.
        Agent occupied slots are synced directly in the DB.

        Returns:
            List of ScheduledSessionData for allocated sessions
        """
        return await self._db_source.allocate_sessions(allocation_batch)

    @scheduler_repository_resilience.apply()
    async def get_pending_timeout_sessions(self) -> list[SweptSessionInfo]:
        """
        Get sessions that have exceeded their pending timeout.
        The timeout is determined by each scaling group's scheduler_opts.
        """
        return await self._db_source.get_pending_timeout_sessions()

    @scheduler_repository_resilience.apply()
    async def batch_update_terminated_status(
        self,
        session_results: list[SessionTerminationResult],
    ) -> None:
        """
        Update terminated status in DB.
        Agent occupied slots are synced directly in the DB.
        """
        if not session_results:
            return

        await self._db_source.batch_update_terminated_status(session_results)

    @scheduler_repository_resilience.apply()
    async def batch_update_kernels_terminated(
        self,
        kernel_results: list[KernelTerminationResult],
        reason: str,
    ) -> None:
        """
        Update kernel statuses to TERMINATED without updating session status.
        Agent occupied slots are synced directly in the DB.
        """
        if not kernel_results:
            return

        await self._db_source.batch_update_kernels_terminated(kernel_results, reason)

    @scheduler_repository_resilience.apply()
    async def mark_sessions_terminating(
        self, session_ids: list[SessionId], reason: str = "USER_REQUESTED"
    ) -> MarkTerminatingResult:
        """
        Mark sessions for termination.
        """
        # Delegate to DB source
        return await self._db_source.mark_sessions_terminating(session_ids, reason)

    @scheduler_repository_resilience.apply()
    async def get_schedulable_scaling_groups(self) -> list[str]:
        """
        Get list of scaling groups that have schedulable agents.
        For sokovan scheduler compatibility.
        """
        return await self._db_source.get_schedulable_scaling_groups()

    @scheduler_repository_resilience.apply()
    async def get_terminating_sessions(self) -> list[TerminatingSessionData]:
        """
        Get sessions with TERMINATING status.
        For sokovan scheduler compatibility.
        """
        return await self._db_source.get_terminating_sessions()

    @scheduler_repository_resilience.apply()
    async def get_terminating_kernels_with_lost_agents(
        self,
    ) -> list[TerminatingKernelWithAgentData]:
        """
        Get kernels in TERMINATING sessions that have lost or missing agents.
        For lost agent cleanup operations.
        """
        return await self._db_source.get_terminating_kernels_with_lost_agents()

    async def _get_known_slot_types(self) -> Mapping[SlotName, SlotTypes]:
        """
        Get known slot types from configuration.
        """
        return await self._config_provider.legacy_etcd_config_loader.get_resource_slots()

    async def _get_max_container_count(self) -> Optional[int]:
        """
        Get max container count from configuration.
        """
        raw_value = await self._config_provider.legacy_etcd_config_loader.get_raw(
            "config/agent/max-container-count"
        )
        return int(raw_value) if raw_value else None

    @scheduler_repository_resilience.apply()
    async def enqueue_session(
        self,
        session_data: SessionEnqueueData,
    ) -> SessionId:
        """
        Enqueue a new session with its kernels.

        Args:
            session_data: Prepared session data with kernels and dependencies

        Returns:
            SessionId: The ID of the created session
        """
        return await self._db_source.enqueue_session(session_data)

    @scheduler_repository_resilience.apply()
    async def fetch_session_creation_data(
        self,
        spec: SessionCreationSpec,
        scaling_group_name: str,
        storage_manager,
        allowed_vfolder_types: list[str],
    ) -> SessionCreationContext:
        """
        Fetch all data needed for session creation in a single DB session.

        Args:
            spec: Session creation specification
            scaling_group_name: Name of the scaling group
            storage_manager: Storage session manager
            allowed_vfolder_types: Allowed vfolder types from config

        Returns:
            SessionCreationContext with all required data
        """
        return await self._db_source.fetch_session_creation_data(
            spec, scaling_group_name, storage_manager, allowed_vfolder_types
        )

    @scheduler_repository_resilience.apply()
    async def fetch_session_creation_context(
        self,
        spec: SessionCreationSpec,
        scaling_group_name: str,
    ) -> SessionCreationContext:
        """
        Legacy method for backward compatibility.
        Use fetch_session_creation_data instead.

        Args:
            spec: Session creation specification
            scaling_group_name: Name of the scaling group

        Returns:
            SessionCreationContext with all required data
        """
        return await self._db_source.fetch_session_creation_context(spec, scaling_group_name)

    @scheduler_repository_resilience.apply()
    async def query_allowed_scaling_groups(
        self,
        domain_name: str,
        group_id: str,
        access_key: str,
    ) -> list[AllowedScalingGroup]:
        """
        Query allowed scaling groups for a user.

        Args:
            domain_name: Domain name
            group_id: Group ID
            access_key: Access key

        Returns:
            List of AllowedScalingGroup objects
        """
        return await self._db_source.query_allowed_scaling_groups(domain_name, group_id, access_key)

    @scheduler_repository_resilience.apply()
    async def prepare_vfolder_mounts(
        self,
        storage_manager,
        allowed_vfolder_types: list[str],
        user_scope,
        resource_policy: dict,
        combined_mounts: list,
        combined_mount_map: dict,
        requested_mount_options: dict,
    ) -> list[VFolderMount]:
        """
        Prepare vfolder mounts for the session.
        """
        return await self._db_source.prepare_vfolder_mounts(
            storage_manager,
            allowed_vfolder_types,
            user_scope,
            resource_policy,
            combined_mounts,
            combined_mount_map,
            requested_mount_options,
        )

    @scheduler_repository_resilience.apply()
    async def prepare_dotfiles(
        self,
        user_scope,
        access_key: AccessKey,
        vfolder_mounts: list[VFolderMount],
    ) -> dict:
        """
        Prepare dotfile data for the session.
        """
        return await self._db_source.prepare_dotfiles(
            user_scope,
            access_key,
            vfolder_mounts,
        )

    @scheduler_repository_resilience.apply()
    async def check_available_image(
        self, image_identifier: ImageIdentifier, domain: str, user_uuid: UUID
    ) -> None:
        """
        Check if the specified image is available.
        Raises ImageNotFound if the image does not exist.
        """
        await self._db_source.check_available_image(image_identifier, domain, user_uuid)

    @scheduler_repository_resilience.apply()
    async def update_sessions_to_prepared(self, session_ids: list[SessionId]) -> None:
        """
        Update sessions from PULLING or PREPARING to PREPARED state.
        """
        await self._db_source.update_sessions_to_prepared(session_ids)

    @scheduler_repository_resilience.apply()
    async def get_sessions_ready_to_run(self) -> list[SessionId]:
        """
        Get sessions in CREATING state where all kernels are RUNNING.
        These sessions are ready to transition to RUNNING state.
        """
        return await self._db_source.get_sessions_ready_to_run()

    @scheduler_repository_resilience.apply()
    async def get_sessions_for_transition(
        self,
        session_statuses: list[SessionStatus],
        kernel_statuses: list[KernelStatus],
    ) -> list[SessionTransitionData]:
        """
        Get sessions ready for state transition based on current session and kernel status.

        :param session_statuses: List of current session statuses to filter by
        :param kernel_statuses: List of current kernel statuses to filter by
        :return: List of sessions ready for transition with detailed information
        """
        return await self._db_source.get_sessions_for_transition(session_statuses, kernel_statuses)

    @scheduler_repository_resilience.apply()
    async def update_sessions_to_running(self, sessions_data: list[SessionRunningData]) -> None:
        """
        Update sessions from CREATING to RUNNING state with occupying_slots.
        """
        await self._db_source.update_sessions_to_running(sessions_data)

    @scheduler_repository_resilience.apply()
    async def get_sessions_ready_to_terminate(self) -> list[SessionId]:
        """
        Get sessions in TERMINATING state where all kernels are TERMINATED.
        These sessions are ready to transition to TERMINATED state.
        """
        return await self._db_source.get_sessions_ready_to_terminate()

    @scheduler_repository_resilience.apply()
    async def update_sessions_to_terminated(self, session_ids: list[SessionId]) -> None:
        """
        Update sessions from TERMINATING to TERMINATED state.
        """
        await self._db_source.update_sessions_to_terminated(session_ids)

    @scheduler_repository_resilience.apply()
    async def update_kernels_to_pulling_for_image(
        self, agent_id: AgentId, image: str, image_ref: Optional[str] = None
    ) -> int:
        """
        Update kernel status from PREPARING to PULLING for the specified image on an agent.

        :param agent_id: The agent ID where kernels should be updated
        :param image: The image name to match kernels
        :param image_ref: Optional image reference
        :return: Number of kernels updated
        """
        return await self._db_source.update_kernels_to_pulling_for_image(agent_id, image, image_ref)

    @scheduler_repository_resilience.apply()
    async def update_kernels_to_prepared_for_image(
        self, agent_id: AgentId, image: str, image_ref: Optional[str] = None
    ) -> int:
        """
        Update kernel status to PREPARED for the specified image on an agent.
        Updates kernels in both PULLING and PREPARING states.

        :param agent_id: The agent ID where kernels should be updated
        :param image: The image name to match kernels
        :param image_ref: Optional image reference
        :return: Number of kernels updated
        """
        return await self._db_source.update_kernels_to_prepared_for_image(
            agent_id, image, image_ref
        )

    @scheduler_repository_resilience.apply()
    async def cancel_kernels_for_failed_image(
        self, agent_id: AgentId, image: str, error_msg: str, image_ref: Optional[str] = None
    ) -> set[SessionId]:
        """
        Cancel kernels for an image that failed to be available on an agent.
        Also checks and cancels sessions if all their kernels are cancelled.

        :param agent_id: The agent ID where the image is unavailable
        :param image: The image name that failed
        :param error_msg: The error message to include in status
        :param image_ref: Optional image reference
        :return: Set of affected session IDs
        """
        affected_session_ids = await self._db_source.cancel_kernels_for_failed_image(
            agent_id, image, error_msg, image_ref
        )

        # Check if any sessions need to be cancelled
        for session_id in affected_session_ids:
            await self._db_source.check_and_cancel_session_if_needed(session_id)

        return affected_session_ids

    @scheduler_repository_resilience.apply()
    async def check_and_cancel_session_if_needed(self, session_id: SessionId) -> bool:
        """
        Check if a session should be cancelled when all its kernels are cancelled.

        :param session_id: The session ID to check
        :return: True if session was cancelled, False otherwise
        """
        return await self._db_source.check_and_cancel_session_if_needed(session_id)

    @scheduler_repository_resilience.apply()
    async def update_kernel_status_pulling(self, kernel_id: UUID, reason: str) -> bool:
        """Update kernel status to PULLING."""
        return await self._db_source.update_kernel_status_pulling(kernel_id, reason)

    @scheduler_repository_resilience.apply()
    async def update_kernel_status_creating(self, kernel_id: UUID, reason: str) -> bool:
        """Update kernel status to CREATING."""
        return await self._db_source.update_kernel_status_creating(kernel_id, reason)

    @scheduler_repository_resilience.apply()
    async def update_kernel_status_running(
        self, kernel_id: UUID, reason: str, creation_info: KernelCreationInfo
    ) -> bool:
        """Update kernel status to RUNNING."""
        return await self._db_source.update_kernel_status_running(kernel_id, reason, creation_info)

    @scheduler_repository_resilience.apply()
    async def update_kernel_status_preparing(self, kernel_id: UUID) -> bool:
        """Update kernel status to PREPARING."""
        return await self._db_source.update_kernel_status_preparing(kernel_id)

    @scheduler_repository_resilience.apply()
    async def update_kernel_status_cancelled(self, kernel_id: UUID, reason: str) -> bool:
        """Update kernel status to CANCELLED."""
        return await self._db_source.update_kernel_status_cancelled(kernel_id, reason)

    @scheduler_repository_resilience.apply()
    async def update_kernel_status_terminated(
        self, kernel_id: UUID, reason: str, exit_code: Optional[int] = None
    ) -> bool:
        """Update kernel status to TERMINATED."""
        return await self._db_source.update_kernel_status_terminated(kernel_id, reason, exit_code)

    @scheduler_repository_resilience.apply()
    async def update_kernels_to_terminated(self, kernel_ids: list[str], reason: str) -> int:
        """Update multiple kernels to TERMINATED status."""
        return await self._db_source.update_kernels_to_terminated(kernel_ids, reason)

    @scheduler_repository_resilience.apply()
    async def update_kernel_heartbeat(self, kernel_id: UUID) -> bool:
        """Update kernel heartbeat timestamp."""
        return await self._db_source.update_kernel_heartbeat(kernel_id)

    @scheduler_repository_resilience.apply()
    async def get_sessions_for_pull(
        self,
        statuses: list[SessionStatus],
        kernel_statuses: list[KernelStatus],
    ) -> SessionsForPullWithImages:
        """
        Get sessions for image pulling with specified statuses.
        Returns SessionsForPullWithImages dataclass.

        :param statuses: Session statuses to filter by (typically SCHEDULED, PREPARING, PULLING)
        :param kernel_statuses: Kernel statuses to filter by (typically PREPARED, PULLING)
        :return: SessionsForPullWithImages object
        """
        return await self._db_source.get_sessions_for_pull(statuses, kernel_statuses)

    @scheduler_repository_resilience.apply()
    async def get_sessions_for_start(
        self,
        session_statuses: list[SessionStatus],
        kernel_statuses: list[KernelStatus],
    ) -> SessionsForStartWithImages:
        """
        Get sessions for starting with specified statuses.
        Returns SessionsForStartWithImages dataclass.

        :param statuses: Session statuses to filter by (typically PREPARED, CREATING)
        :param kernel_statuses: Kernel statuses to filter by (typically PREPARED)
        :return: SessionsForStartWithImages object
        """
        return await self._db_source.get_sessions_for_start(session_statuses, kernel_statuses)

    @scheduler_repository_resilience.apply()
    async def update_sessions_to_preparing(self, session_ids: list[SessionId]) -> None:
        """
        Update sessions from SCHEDULED to PREPARING status.
        """
        await self._db_source.update_sessions_to_preparing(session_ids)

    @scheduler_repository_resilience.apply()
    async def update_sessions_and_kernels_to_creating(self, session_ids: list[SessionId]) -> None:
        """
        Update sessions and kernels from PREPARED to CREATING status.
        """
        await self._db_source.update_sessions_and_kernels_to_creating(session_ids)

    @scheduler_repository_resilience.apply()
    async def mark_session_cancelled(
        self, session_id: SessionId, error_info: ErrorStatusInfo, reason: str = "FAILED_TO_START"
    ) -> None:
        """
        Mark a session as cancelled with error information.
        """
        await self._db_source.mark_session_cancelled(session_id, error_info, reason)

    @scheduler_repository_resilience.apply()
    async def get_container_info_for_kernels(
        self, session_id: SessionId
    ) -> dict[UUID, Optional[str]]:
        """
        Get container IDs for kernels in a session.
        """
        return await self._db_source.get_container_info_for_kernels(session_id)

    @scheduler_repository_resilience.apply()
    async def batch_update_stuck_session_retries(
        self, session_ids: list[SessionId], max_retries: int = 5
    ) -> list[SessionId]:
        """
        Batch update retry counts for stuck sessions.
        Sessions that exceed max_retries are moved to PENDING status.

        :param session_ids: List of session IDs to update
        :param max_retries: Maximum retries before moving to PENDING (default: 5)
        :return: List of session IDs that should continue retrying (not moved to PENDING)
        """
        return await self._db_source.batch_update_stuck_session_retries(session_ids, max_retries)

    @scheduler_repository_resilience.apply()
    async def update_session_error_info(
        self, session_id: SessionId, error_info: ErrorStatusInfo
    ) -> None:
        """
        Update session's status_data with error information without changing status.
        This is used when a session fails but should be retried later.

        :param session_id: The session ID to update
        :param error_info: Error information to store in status_data
        """
        await self._db_source.update_session_error_info(session_id, error_info)

    @scheduler_repository_resilience.apply()
    async def get_keypair_concurrency(self, access_key: AccessKey, is_sftp: bool = False) -> int:
        """
        Get keypair concurrency with cache-through pattern.
        First checks cache, falls back to DB if not cached, and updates cache.

        :param access_key: The access key to query
        :param is_sftp: Whether to get SFTP concurrency (True) or regular concurrency (False)
        :return: Current concurrency count
        """
        # Try to get from cache first
        try:
            cached_value = await self._cache_source.get_keypair_concurrency(access_key, is_sftp)
            if cached_value is not None:
                return cached_value
        except Exception as e:
            log.warning(
                "Failed to get keypair concurrency from cache for {}: {}",
                access_key,
                e,
            )

        # Cache miss - refresh both values from DB
        concurrency_data = await self._db_source.get_keypair_concurrencies_from_db(access_key)

        # Update cache with both values at once
        try:
            await self._cache_source.set_keypair_concurrencies(
                access_key,
                concurrency_data.regular_count,
                concurrency_data.sftp_count,
            )
        except Exception as e:
            log.warning(
                "Failed to update keypair concurrency cache for {}: {}",
                access_key,
                e,
            )

        # Return the requested value
        return concurrency_data.sftp_count if is_sftp else concurrency_data.regular_count

    @scheduler_repository_resilience.apply()
    async def invalidate_kernel_related_cache(self, access_keys: list[AccessKey]) -> None:
        """
        Invalidate caches related to kernel state changes affecting resource calculations.

        :param access_keys: List of access keys whose related caches should be invalidated
        """
        await self._cache_source.invalidate_kernel_related_cache(access_keys)

    @scheduler_repository_resilience.apply()
    async def update_session_network_id(
        self,
        session_id: SessionId,
        network_id: Optional[str],
    ) -> None:
        """
        Update the network ID associated with a session.
        :param session_id: The session ID to update
        :param network_id: The new network ID to set (or None to clear)
        """
        await self._db_source.update_session_network_id(
            session_id,
            network_id,
        )

    @scheduler_repository_resilience.apply()
    async def get_total_resource_slots(self) -> TotalResourceData:
        """
        Get total resource slots from all agents.
        Uses cache-through pattern: checks cache first, falls back to DB calculation,
        and updates cache.

        :return: TotalResourceData with total used, free, and capable slots
        """
        # Try to get from cache first
        try:
            cached_data = await self._cache_source.get_total_resource_slots()
            if cached_data is not None:
                return cached_data
        except Exception as e:
            log.warning("Failed to get total resource slots from cache: {}", e)

        # Cache miss - calculate from DB
        total_resource_data = await self._db_source.calculate_total_resource_slots()

        # Update cache with calculated value
        try:
            await self._cache_source.set_total_resource_slots(total_resource_data)
        except Exception as e:
            log.warning("Failed to update total resource slots cache: {}", e)

        return total_resource_data



from collections.abc import Mapping
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.common.types import ImageID
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.image.types import (
    ImageStatus,
    ImageWithAgentInstallStatus,
)
from ai.backend.manager.services.image.actions.base import ImageAction


@dataclass
class GetAllImagesAction(ImageAction):
    """
    Action to retrieve all images, optionally filtered by their status.
    Args:
        status_filter: If provided, only images with a status in this list will be returned.
            If None, images of all statuses will be included.
    """

    status_filter: Optional[list[ImageStatus]]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "get_all_images"


@dataclass
class GetAllImagesActionResult(BaseActionResult):
    data: Mapping[ImageID, ImageWithAgentInstallStatus]

    @override
    def entity_id(self) -> Optional[str]:
        return None



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.artifact_revision.actions.approve import (
    ApproveArtifactRevisionAction,
    ApproveArtifactRevisionActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.associate_with_storage import (
    AssociateWithStorageAction,
    AssociateWithStorageActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.cancel_import import (
    CancelImportAction,
    CancelImportActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.cleanup import (
    CleanupArtifactRevisionAction,
    CleanupArtifactRevisionActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.delegate_import_revision_batch import (
    DelegateImportArtifactRevisionBatchAction,
    DelegateImportArtifactRevisionBatchActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.disassociate_with_storage import (
    DisassociateWithStorageAction,
    DisassociateWithStorageActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.get import (
    GetArtifactRevisionAction,
    GetArtifactRevisionActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.get_download_progress import (
    GetDownloadProgressAction,
    GetDownloadProgressActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.get_readme import (
    GetArtifactRevisionReadmeAction,
    GetArtifactRevisionReadmeActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.import_revision import (
    ImportArtifactRevisionAction,
    ImportArtifactRevisionActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.list import (
    ListArtifactRevisionsAction,
    ListArtifactRevisionsActionResult,
)
from ai.backend.manager.services.artifact_revision.actions.reject import (
    RejectArtifactRevisionAction,
    RejectArtifactRevisionActionResult,
)
from ai.backend.manager.services.artifact_revision.service import ArtifactRevisionService


class ArtifactRevisionProcessors(AbstractProcessorPackage):
    get: ActionProcessor[GetArtifactRevisionAction, GetArtifactRevisionActionResult]
    get_readme: ActionProcessor[
        GetArtifactRevisionReadmeAction, GetArtifactRevisionReadmeActionResult
    ]
    get_download_progress: ActionProcessor[
        GetDownloadProgressAction, GetDownloadProgressActionResult
    ]
    list_revision: ActionProcessor[ListArtifactRevisionsAction, ListArtifactRevisionsActionResult]
    approve: ActionProcessor[ApproveArtifactRevisionAction, ApproveArtifactRevisionActionResult]
    reject: ActionProcessor[RejectArtifactRevisionAction, RejectArtifactRevisionActionResult]
    import_revision: ActionProcessor[
        ImportArtifactRevisionAction, ImportArtifactRevisionActionResult
    ]
    delegate_import_revision_batch: ActionProcessor[
        DelegateImportArtifactRevisionBatchAction, DelegateImportArtifactRevisionBatchActionResult
    ]
    cancel_import: ActionProcessor[CancelImportAction, CancelImportActionResult]
    cleanup: ActionProcessor[CleanupArtifactRevisionAction, CleanupArtifactRevisionActionResult]
    associate_with_storage: ActionProcessor[
        AssociateWithStorageAction, AssociateWithStorageActionResult
    ]
    disassociate_with_storage: ActionProcessor[
        DisassociateWithStorageAction, DisassociateWithStorageActionResult
    ]

    def __init__(
        self, service: ArtifactRevisionService, action_monitors: list[ActionMonitor]
    ) -> None:
        self.get = ActionProcessor(service.get, action_monitors)
        self.get_readme = ActionProcessor(service.get_readme, action_monitors)
        self.get_download_progress = ActionProcessor(service.get_download_progress, action_monitors)
        self.list_revision = ActionProcessor(service.list_revision, action_monitors)
        self.approve = ActionProcessor(service.approve, action_monitors)
        self.reject = ActionProcessor(service.reject, action_monitors)
        self.import_revision = ActionProcessor(service.import_revision, action_monitors)
        self.delegate_import_revision_batch = ActionProcessor(
            service.delegate_import_revision_batch, action_monitors
        )
        self.cancel_import = ActionProcessor(service.cancel_import, action_monitors)
        self.cleanup = ActionProcessor(service.cleanup, action_monitors)
        self.associate_with_storage = ActionProcessor(
            service.associate_with_storage, action_monitors
        )
        self.disassociate_with_storage = ActionProcessor(
            service.disassociate_with_storage, action_monitors
        )

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            GetArtifactRevisionAction.spec(),
            GetArtifactRevisionReadmeAction.spec(),
            GetDownloadProgressAction.spec(),
            ListArtifactRevisionsAction.spec(),
            ApproveArtifactRevisionAction.spec(),
            RejectArtifactRevisionAction.spec(),
            ImportArtifactRevisionAction.spec(),
            CancelImportAction.spec(),
            CleanupArtifactRevisionAction.spec(),
            AssociateWithStorageAction.spec(),
            DisassociateWithStorageAction.spec(),
            DelegateImportArtifactRevisionBatchAction.spec(),
        ]



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.deployment.types import ModelDeploymentAccessTokenData
from ai.backend.manager.repositories.deployment.types.types import AccessTokenOrderingOptions
from ai.backend.manager.services.deployment.actions.base import DeploymentBaseAction
from ai.backend.manager.types import PaginationOptions


@dataclass
class ListAccessTokensAction(DeploymentBaseAction):
    pagination: PaginationOptions
    ordering: Optional[AccessTokenOrderingOptions] = None

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "list_access_tokens"


@dataclass
class ListAccessTokensActionResult(BaseActionResult):
    data: list[ModelDeploymentAccessTokenData]
    total_count: int

    @override
    def entity_id(self) -> Optional[str]:
        return None



from typing import Any
from uuid import UUID

import aiohttp

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience

from .types import CreateEndpointRequestBody

appproxy_client_resilience = Resilience(
    policies=[
        MetricPolicy(MetricArgs(domain=DomainType.CLIENT, layer=LayerType.WSPROXY_CLIENT)),
        RetryPolicy(
            RetryArgs(
                max_retries=3,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.EXPONENTIAL,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class AppProxyClient:
    _client_session: aiohttp.ClientSession
    _address: str
    _token: str

    def __init__(self, client_session: aiohttp.ClientSession, address: str, token: str) -> None:
        self._client_session = client_session
        self._address = address
        self._token = token

    @appproxy_client_resilience.apply()
    async def create_endpoint(
        self,
        endpoint_id: UUID,
        body: CreateEndpointRequestBody,
    ) -> dict[str, Any]:
        async with self._client_session.post(
            f"/v2/endpoints/{endpoint_id}",
            json=body.model_dump(mode="json"),
            headers={
                "X-BackendAI-Token": self._token,
            },
        ) as resp:
            resp.raise_for_status()
            return await resp.json()

    @appproxy_client_resilience.apply()
    async def delete_endpoint(
        self,
        endpoint_id: UUID,
    ) -> None:
        async with self._client_session.delete(
            f"/v2/endpoints/{endpoint_id}",
            headers={
                "X-BackendAI-Token": self._token,
            },
        ):
            pass



from collections.abc import Mapping, Sequence
from datetime import datetime
from enum import StrEnum
from pathlib import PurePosixPath
from typing import Any, Optional, cast
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID, PageInfo
from strawberry.scalars import JSON

from ai.backend.common.exception import ModelDeploymentUnavailable, ModelRevisionNotFound
from ai.backend.common.types import ClusterMode as CommonClusterMode
from ai.backend.common.types import MountPermission as CommonMountPermission
from ai.backend.common.types import RuntimeVariant
from ai.backend.manager.api.gql.base import (
    JSONString,
    OrderDirection,
    StringFilter,
    build_page_info,
    build_pagination_options,
    resolve_global_id,
    to_global_id,
)
from ai.backend.manager.api.gql.image import (
    Image,
)
from ai.backend.manager.api.gql.resource_group import (
    ResourceGroup,
)
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql.vfolder import (
    ExtraVFolderMountConnection,
    VFolder,
)
from ai.backend.manager.data.deployment.creator import ModelRevisionCreator, VFolderMountsCreator
from ai.backend.manager.data.deployment.inference_runtime_config import (
    MOJORuntimeConfig,
    NVDIANIMRuntimeConfig,
    SGLangRuntimeConfig,
    VLLMRuntimeConfig,
)
from ai.backend.manager.data.deployment.types import (
    ClusterConfigData,
    ExecutionSpec,
    ModelMountConfigData,
    ModelRevisionData,
    ModelRevisionOrderField,
    ModelRuntimeConfigData,
    MountInfo,
    ResourceConfigData,
    ResourceSpec,
)
from ai.backend.manager.data.image.types import ImageIdentifier
from ai.backend.manager.models.gql_models.image import ImageNode
from ai.backend.manager.models.gql_models.scaling_group import ScalingGroupNode
from ai.backend.manager.models.gql_models.vfolder import VirtualFolderNode
from ai.backend.manager.repositories.deployment.types.types import (
    ModelRevisionFilterOptions,
    ModelRevisionOrderingOptions,
)
from ai.backend.manager.services.deployment.actions.model_revision.add_model_revision import (
    AddModelRevisionAction,
)
from ai.backend.manager.services.deployment.actions.model_revision.batch_load_revisions import (
    BatchLoadRevisionsAction,
)
from ai.backend.manager.services.deployment.actions.model_revision.create_model_revision import (
    CreateModelRevisionAction,
)
from ai.backend.manager.services.deployment.actions.model_revision.list_revisions import (
    ListRevisionsAction,
)

MountPermission = strawberry.enum(
    CommonMountPermission,
    name="MountPermission",
    description="Added in 25.16.0. This enum represents the permission level for a mounted volume. It can be ro, rw, wd",
)


@strawberry.enum(description="Added in 25.16.0")
class ClusterMode(StrEnum):
    SINGLE_NODE = "SINGLE_NODE"
    MULTI_NODE = "MULTI_NODE"


@strawberry.type(description="Added in 25.16.0")
class ModelMountConfig:
    _vfolder_id: strawberry.Private[UUID]
    mount_destination: str
    definition_path: str

    @strawberry.field
    async def vfolder(self, info: Info[StrawberryGQLContext]) -> VFolder:
        vfolder_global_id = to_global_id(
            VirtualFolderNode, self._vfolder_id, is_target_graphene_object=True
        )
        return VFolder(id=ID(vfolder_global_id))

    @classmethod
    def from_dataclass(cls, data: ModelMountConfigData) -> "ModelMountConfig":
        return cls(
            _vfolder_id=data.vfolder_id,
            mount_destination=data.mount_destination,
            definition_path=data.definition_path,
        )


@strawberry.type(description="Added in 25.16.0")
class ModelRuntimeConfig:
    runtime_variant: str
    inference_runtime_config: Optional[JSON] = None
    environ: Optional[JSONString] = strawberry.field(
        description='Environment variables for the service, e.g. {"CUDA_VISIBLE_DEVICES": "0"}',
        default=None,
    )

    @classmethod
    def from_dataclass(cls, data: ModelRuntimeConfigData) -> "ModelRuntimeConfig":
        return cls(
            runtime_variant=data.runtime_variant,
            inference_runtime_config=data.inference_runtime_config,
            environ=JSONString.serialize(data.environ) if data.environ else None,
        )


@strawberry.type(description="Added in 25.16.0")
class ResourceConfig:
    _resource_group_name: strawberry.Private[str]
    resource_slots: JSONString = strawberry.field(
        description='Resource Slots are a JSON string that describes the resources allocated for the deployment. Example: "resourceSlots": "{\\"cpu\\": \\"1\\", \\"mem\\": \\"1073741824\\", \\"cuda.device\\": \\"0\\"}"'
    )
    resource_opts: Optional[JSONString] = strawberry.field(
        description='Resource Options are a JSON string that describes additional options for the resources. This is especially used for shared memory configurations. Example: "resourceOpts": "{\\"shmem\\": \\"64m\\"}"',
        default=None,
    )

    @strawberry.field
    def resource_group(self) -> "ResourceGroup":
        """Resolves the federated ResourceGroup."""
        global_id = to_global_id(
            ScalingGroupNode, self._resource_group_name, is_target_graphene_object=True
        )
        return ResourceGroup(id=ID(global_id))

    @classmethod
    def from_dataclass(cls, data: ResourceConfigData) -> "ResourceConfig":
        return cls(
            _resource_group_name=data.resource_group_name,
            resource_slots=JSONString.from_resource_slot(data.resource_slot),
            resource_opts=JSONString.serialize(data.resource_opts),
        )


@strawberry.type(description="Added in 25.16.0")
class ClusterConfig:
    mode: ClusterMode
    size: int

    @classmethod
    def from_dataclass(cls, data: ClusterConfigData) -> "ClusterConfig":
        return cls(
            mode=ClusterMode(data.mode.name),
            size=data.size,
        )


@strawberry.type(description="Added in 25.16.0")
class ModelRevision(Node):
    _image_id: strawberry.Private[UUID]
    id: NodeID
    name: str
    cluster_config: ClusterConfig
    resource_config: ResourceConfig
    model_runtime_config: ModelRuntimeConfig
    model_mount_config: ModelMountConfig
    extra_mounts: ExtraVFolderMountConnection
    created_at: datetime

    @strawberry.field
    async def image(self, info: Info[StrawberryGQLContext]) -> Image:
        image_global_id = to_global_id(ImageNode, self._image_id, is_target_graphene_object=True)
        return Image(id=ID(image_global_id))

    @classmethod
    def from_dataclass(cls, data: ModelRevisionData) -> "ModelRevision":
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            cluster_config=ClusterConfig.from_dataclass(data.cluster_config),
            resource_config=ResourceConfig.from_dataclass(data.resource_config),
            model_runtime_config=ModelRuntimeConfig.from_dataclass(data.model_runtime_config),
            model_mount_config=ModelMountConfig.from_dataclass(data.model_mount_config),
            extra_mounts=ExtraVFolderMountConnection.from_dataclass(data.extra_vfolder_mounts),
            _image_id=data.image_id,
            created_at=data.created_at,
        )

    @classmethod
    async def batch_load_by_ids(
        cls, ctx: StrawberryGQLContext, revision_ids: Sequence[UUID]
    ) -> list["ModelRevision"]:
        """Batch load revisions by their IDs."""
        processor = ctx.processors.deployment
        if processor is None:
            raise ModelDeploymentUnavailable(
                "Model Deployment feature is unavailable. Please contact support."
            )

        result = await processor.batch_load_revisions.wait_for_complete(
            BatchLoadRevisionsAction(revision_ids=list(revision_ids))
        )

        revision_map = {revision.id: revision for revision in result.data}
        revisions = []
        for revision_id in revision_ids:
            if revision_id not in revision_map:
                raise ModelRevisionNotFound(f"Revision {revision_id} not found")
            revisions.append(cls.from_dataclass(revision_map[revision_id]))

        return revisions


# Filter and Order Types
@strawberry.input(description="Added in 25.16.0")
class ModelRevisionFilter:
    name: Optional[StringFilter] = None
    deployment_id: Optional[ID] = None
    id: Optional[ID] = None
    ids_in: strawberry.Private[Optional[Sequence[UUID]]] = None

    AND: Optional[list["ModelRevisionFilter"]] = None
    OR: Optional[list["ModelRevisionFilter"]] = None
    NOT: Optional[list["ModelRevisionFilter"]] = None

    def to_repo_filter(self) -> ModelRevisionFilterOptions:
        repo_filter = ModelRevisionFilterOptions()

        # Handle basic filters
        repo_filter.name = self.name
        repo_filter.deployment_id = UUID(self.deployment_id) if self.deployment_id else None
        repo_filter.id = UUID(self.id) if self.id else None
        repo_filter.ids_in = list(self.ids_in) if self.ids_in is not None else None

        # Handle logical operations
        if self.AND:
            repo_filter.AND = [f.to_repo_filter() for f in self.AND]
        if self.OR:
            repo_filter.OR = [f.to_repo_filter() for f in self.OR]
        if self.NOT:
            repo_filter.NOT = [f.to_repo_filter() for f in self.NOT]

        return repo_filter


@strawberry.input(description="Added in 25.16.0")
class ModelRevisionOrderBy:
    field: ModelRevisionOrderField
    direction: OrderDirection = OrderDirection.DESC


# Payload Types
@strawberry.type(description="Added in 25.16.0")
class CreateModelRevisionPayload:
    revision: ModelRevision


@strawberry.type(description="Added in 25.16.0")
class AddModelRevisionPayload:
    revision: ModelRevision


# Input Types
@strawberry.input(description="Added in 25.16.0")
class ClusterConfigInput:
    mode: ClusterMode
    size: int


@strawberry.input(description="Added in 25.16.0")
class ResourceGroupInput:
    name: str


@strawberry.input(description="Added in 25.16.0")
class ResourceConfigInput:
    resource_group: ResourceGroupInput
    resource_slots: JSONString = strawberry.field(
        description='Resources allocated for the deployment. Example: "resourceSlots": "{\\"cpu\\": \\"1\\", \\"mem\\": \\"1073741824\\", \\"cuda.device\\": \\"0\\"}"'
    )
    resource_opts: Optional[JSONString] = strawberry.field(
        description='Additional options for the resources. This is especially used for shared memory configurations. Example: "resourceOpts": "{\\"shmem\\": \\"64m\\"}"',
        default=None,
    )


@strawberry.input(description="Added in 25.16.0")
class ImageInput:
    name: str
    architecture: str


@strawberry.input(description="Added in 25.16.0")
class ModelRuntimeConfigInput:
    runtime_variant: str
    inference_runtime_config: Optional[JSON] = None
    environ: Optional[JSONString] = strawberry.field(
        description='Environment variables for the service, e.g. {"CUDA_VISIBLE_DEVICES": "0"}',
        default=None,
    )


@strawberry.input(description="Added in 25.16.0")
class ModelMountConfigInput:
    vfolder_id: ID
    mount_destination: str
    definition_path: str


@strawberry.input(description="Added in 25.16.0")
class ExtraVFolderMountInput:
    vfolder_id: ID
    mount_destination: Optional[str]


@strawberry.input(description="Added in 25.16.0")
class CreateModelRevisionInput:
    name: Optional[str] = None
    cluster_config: ClusterConfigInput
    resource_config: ResourceConfigInput
    image: ImageInput
    model_runtime_config: ModelRuntimeConfigInput
    model_mount_config: ModelMountConfigInput
    extra_mounts: Optional[list[ExtraVFolderMountInput]]

    def to_model_revision_creator(self) -> ModelRevisionCreator:
        image_identifier = ImageIdentifier(
            canonical=self.image.name,
            architecture=self.image.architecture,
        )

        resource_spec = ResourceSpec(
            cluster_mode=CommonClusterMode(self.cluster_config.mode),
            cluster_size=self.cluster_config.size,
            resource_slots=cast(Mapping[str, Any], self.resource_config.resource_slots),
            resource_opts=cast(Mapping[str, Any] | None, self.resource_config.resource_opts),
        )

        extra_mounts = []
        if self.extra_mounts is not None:
            extra_mounts = [
                MountInfo(
                    vfolder_id=UUID(str(extra_mount.vfolder_id)),
                    kernel_path=PurePosixPath(
                        extra_mount.mount_destination
                        if extra_mount.mount_destination is not None
                        else ""
                    ),
                )
                for extra_mount in self.extra_mounts
            ]

        mounts = VFolderMountsCreator(
            model_vfolder_id=UUID(str(self.model_mount_config.vfolder_id)),
            model_definition_path=self.model_mount_config.definition_path,
            model_mount_destination=self.model_mount_config.mount_destination,
            extra_mounts=extra_mounts,
        )

        execution_spec = ExecutionSpec(
            environ=cast(Optional[dict[str, str]], self.model_runtime_config.environ),
            runtime_variant=RuntimeVariant(self.model_runtime_config.runtime_variant),
            inference_runtime_config=cast(
                Optional[dict[str, Any]], self.model_runtime_config.inference_runtime_config
            ),
        )

        return ModelRevisionCreator(
            image_identifier=image_identifier,
            resource_spec=resource_spec,
            mounts=mounts,
            execution=execution_spec,
        )


@strawberry.input(description="Added in 25.16.0")
class AddModelRevisionInput:
    name: Optional[str] = None
    deployment_id: ID
    cluster_config: ClusterConfigInput
    resource_config: ResourceConfigInput
    image: ImageInput
    model_runtime_config: ModelRuntimeConfigInput
    model_mount_config: ModelMountConfigInput
    extra_mounts: Optional[list[ExtraVFolderMountInput]]

    def to_model_revision_creator(self) -> ModelRevisionCreator:
        image_identifier = ImageIdentifier(
            canonical=self.image.name,
            architecture=self.image.architecture,
        )

        resource_spec = ResourceSpec(
            cluster_mode=CommonClusterMode(self.cluster_config.mode),
            cluster_size=self.cluster_config.size,
            resource_slots=cast(Mapping[str, Any], self.resource_config.resource_slots),
            resource_opts=cast(Mapping[str, Any] | None, self.resource_config.resource_opts),
        )

        extra_mounts = []
        if self.extra_mounts is not None:
            extra_mounts = [
                MountInfo(
                    vfolder_id=UUID(str(extra_mount.vfolder_id)),
                    kernel_path=PurePosixPath(
                        extra_mount.mount_destination
                        if extra_mount.mount_destination is not None
                        else ""
                    ),
                )
                for extra_mount in self.extra_mounts
            ]

        mounts = VFolderMountsCreator(
            model_vfolder_id=UUID(str(self.model_mount_config.vfolder_id)),
            model_definition_path=self.model_mount_config.definition_path,
            model_mount_destination=self.model_mount_config.mount_destination,
            extra_mounts=extra_mounts,
        )

        execution_spec = ExecutionSpec(
            environ=cast(Optional[dict[str, str]], self.model_runtime_config.environ),
            runtime_variant=RuntimeVariant(self.model_runtime_config.runtime_variant),
            inference_runtime_config=cast(
                Optional[dict[str, Any]], self.model_runtime_config.inference_runtime_config
            ),
        )

        return ModelRevisionCreator(
            image_identifier=image_identifier,
            resource_spec=resource_spec,
            mounts=mounts,
            execution=execution_spec,
        )


ModelRevisionEdge = Edge[ModelRevision]


@strawberry.type(description="Added in 25.16.0")
class ModelRevisionConnection(Connection[ModelRevision]):
    count: int

    def __init__(self, *args, count: int, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.count = count

    @classmethod
    def from_dataclass(cls, revisions_data: list[ModelRevisionData]) -> "ModelRevisionConnection":
        nodes = [ModelRevision.from_dataclass(data) for data in revisions_data]
        edges = [ModelRevisionEdge(node=node, cursor=str(node.id)) for node in nodes]

        page_info = PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        )

        return cls(count=len(nodes), edges=edges, page_info=page_info)


@strawberry.field(
    description="Added in 25.16.0. Get JSON Schema for inference runtime configuration"
)
async def inference_runtime_config(name: str) -> JSON:
    match name.lower():
        case "vllm":
            return VLLMRuntimeConfig.to_json_schema()
        case "sglang":
            return SGLangRuntimeConfig.to_json_schema()
        case "nvdianim":
            return NVDIANIMRuntimeConfig.to_json_schema()
        case "mojo":
            return MOJORuntimeConfig.to_json_schema()
        case _:
            return {
                "error": "Unknown service name",
            }


@strawberry.field(
    description="Added in 25.16.0 Get configuration JSON Schemas for all inference runtimes"
)
async def inference_runtime_configs(info: Info[StrawberryGQLContext]) -> JSON:
    all_configs = {
        "vllm": VLLMRuntimeConfig.to_json_schema(),
        "sglang": SGLangRuntimeConfig.to_json_schema(),
        "nvdianim": NVDIANIMRuntimeConfig.to_json_schema(),
        "mojo": MOJORuntimeConfig.to_json_schema(),
    }

    return all_configs


def _convert_gql_revision_ordering_to_repo_ordering(
    order_by: Optional[list[ModelRevisionOrderBy]],
) -> ModelRevisionOrderingOptions:
    if order_by is None or len(order_by) == 0:
        return ModelRevisionOrderingOptions()

    repo_ordering = []
    for order in order_by:
        desc = order.direction == OrderDirection.DESC
        repo_ordering.append((order.field, desc))

    return ModelRevisionOrderingOptions(order_by=repo_ordering)


async def resolve_revisions(
    info: Info[StrawberryGQLContext],
    filter: Optional[ModelRevisionFilter] = None,
    order_by: Optional[list[ModelRevisionOrderBy]] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> ModelRevisionConnection:
    repo_filter = None
    if filter:
        repo_filter = filter.to_repo_filter()

    repo_ordering = _convert_gql_revision_ordering_to_repo_ordering(order_by)

    pagination_options = build_pagination_options(
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )

    processor = info.context.processors.deployment
    if processor is None:
        raise ModelDeploymentUnavailable(
            "Model Deployment feature is unavailable. Please contact support."
        )
    action_result = await processor.list_revisions.wait_for_complete(
        ListRevisionsAction(
            pagination=pagination_options,
            ordering=repo_ordering,
            filters=repo_filter,
        )
    )

    edges = []
    revisions = action_result.data
    total_count = action_result.total_count
    for revision in revisions:
        edges.append(
            ModelRevisionEdge(
                node=ModelRevision.from_dataclass(revision),
                cursor=to_global_id(ModelRevision, revision.id),
            )
        )

    page_info = build_page_info(edges, total_count, pagination_options)

    connection = ModelRevisionConnection(
        count=total_count,
        edges=edges,
        page_info=page_info.to_strawberry_page_info(),
    )
    return connection


@strawberry.field(description="Added in 25.16.0")
async def revisions(
    info: Info[StrawberryGQLContext],
    filter: Optional[ModelRevisionFilter] = None,
    order_by: Optional[list[ModelRevisionOrderBy]] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> ModelRevisionConnection:
    """List revisions with optional filtering and pagination."""
    return await resolve_revisions(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(description="Added in 25.16.0")
async def revision(id: ID, info: Info[StrawberryGQLContext]) -> ModelRevision:
    """Get a specific revision by ID."""
    _, revision_id = resolve_global_id(id)
    revision_loader = info.context.dataloader_registry.get_loader(
        ModelRevision.batch_load_by_ids, info.context
    )
    revision: list[ModelRevision] = await revision_loader.load(revision_id)
    return revision[0]


@strawberry.mutation(description="Added in 25.16.0")
async def add_model_revision(
    input: AddModelRevisionInput, info: Info[StrawberryGQLContext]
) -> AddModelRevisionPayload:
    """Add a model revision to a deployment."""

    processor = info.context.processors.deployment
    if processor is None:
        raise ModelDeploymentUnavailable(
            "Model Deployment feature is unavailable. Please contact support."
        )

    result = await processor.add_model_revision.wait_for_complete(
        AddModelRevisionAction(
            model_deployment_id=UUID(input.deployment_id), adder=input.to_model_revision_creator()
        )
    )

    return AddModelRevisionPayload(revision=ModelRevision.from_dataclass(result.revision))


@strawberry.mutation(
    description="Added in 25.16.0. Create model revision which is not attached to any deployment."
)
async def create_model_revision(
    input: CreateModelRevisionInput, info: Info[StrawberryGQLContext]
) -> CreateModelRevisionPayload:
    """Create a new model revision without attaching it to any deployment."""
    processor = info.context.processors.deployment
    if processor is None:
        raise ModelDeploymentUnavailable(
            "Model Deployment feature is unavailable. Please contact support."
        )

    result = await processor.create_model_revision.wait_for_complete(
        CreateModelRevisionAction(creator=input.to_model_revision_creator())
    )

    return CreateModelRevisionPayload(revision=ModelRevision.from_dataclass(result.revision))



from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.manager.data.resource.types import KeyPairResourcePolicyData
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.repositories.keypair_resource_policy.db_source.db_source import (
    KeypairResourcePolicyDBSource,
)
from ai.backend.manager.services.keypair_resource_policy.actions.modify_keypair_resource_policy import (
    KeyPairResourcePolicyModifier,
)
from ai.backend.manager.services.keypair_resource_policy.types import KeyPairResourcePolicyCreator

keypair_resource_policy_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(
                domain=DomainType.REPOSITORY, layer=LayerType.KEYPAIR_RESOURCE_POLICY_REPOSITORY
            )
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=5,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class KeypairResourcePolicyRepository:
    _db_source: KeypairResourcePolicyDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = KeypairResourcePolicyDBSource(db)

    @keypair_resource_policy_repository_resilience.apply()
    async def create_keypair_resource_policy(
        self, creator: KeyPairResourcePolicyCreator
    ) -> KeyPairResourcePolicyData:
        return await self._db_source.insert(creator)

    @keypair_resource_policy_repository_resilience.apply()
    async def update_keypair_resource_policy(
        self, name: str, modifier: KeyPairResourcePolicyModifier
    ) -> KeyPairResourcePolicyData:
        return await self._db_source.update(name, modifier)

    @keypair_resource_policy_repository_resilience.apply()
    async def remove_keypair_resource_policy(self, name: str) -> KeyPairResourcePolicyData:
        return await self._db_source.delete(name)



from __future__ import annotations

from ai.backend.manager.data.notification import (
    NotificationChannelCreator,
    NotificationChannelData,
    NotificationChannelModifier,
    NotificationRuleCreator,
    NotificationRuleData,
    NotificationRuleModifier,
)

from .repositories import NotificationRepositories
from .repository import NotificationRepository

__all__ = (
    "NotificationChannelCreator",
    "NotificationChannelData",
    "NotificationChannelModifier",
    "NotificationRepositories",
    "NotificationRepository",
    "NotificationRuleCreator",
    "NotificationRuleData",
    "NotificationRuleModifier",
)



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.common.data.storage.registries.types import ModelTarget
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.artifact.types import ArtifactDataWithRevisions
from ai.backend.manager.services.artifact.actions.base import ArtifactAction


@dataclass
class RetrieveModelsAction(ArtifactAction):
    registry_id: Optional[uuid.UUID]
    models: list[ModelTarget]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "retrieve_model_multi"


@dataclass
class RetrieveModelsActionResult(BaseActionResult):
    result: list[ArtifactDataWithRevisions]

    @override
    def entity_id(self) -> Optional[str]:
        return None



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, override
from uuid import UUID

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.notification import NotificationChannelData

from .base import NotificationAction


@dataclass
class GetChannelAction(NotificationAction):
    """Action to get a notification channel by ID."""

    channel_id: UUID

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "get_channel"

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.channel_id)


@dataclass
class GetChannelActionResult(BaseActionResult):
    """Result of getting a notification channel."""

    channel_data: NotificationChannelData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.channel_data.id)



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.user.types import UserCreateResultData, UserCreator
from ai.backend.manager.services.user.actions.base import UserAction


@dataclass
class CreateUserAction(UserAction):
    creator: UserCreator
    group_ids: Optional[list[str]] = None

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "create"


@dataclass
class CreateUserActionResult(BaseActionResult):
    data: UserCreateResultData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.data.user.id)



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.group.admin_repository import AdminGroupRepository
from ai.backend.manager.repositories.group.repository import GroupRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class GroupRepositories:
    repository: GroupRepository
    admin_repository: AdminGroupRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = GroupRepository(args.db, args.config_provider, args.valkey_stat_client)
        admin_repository = AdminGroupRepository(args.db, args.storage_manager)

        return cls(
            repository=repository,
            admin_repository=admin_repository,
        )



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.artifact.types import ArtifactRevisionData
from ai.backend.manager.services.artifact_revision.actions.base import ArtifactRevisionAction


# TODO: Make this a batch action.
@dataclass
class ImportArtifactBatchAction(ArtifactRevisionAction):
    artifact_revision_ids: list[uuid.UUID]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "import_batch"


@dataclass
class ImportArtifactBatchActionResult(BaseActionResult):
    result: list[ArtifactRevisionData]

    @override
    def entity_id(self) -> Optional[str]:
        return None



from abc import abstractmethod
from typing import Optional, TypeVar, override

from ai.backend.manager.data.permission.types import OperationType

from .base import BaseAction, BaseActionResult


class BaseScopeAction(BaseAction):
    @override
    def entity_id(self) -> Optional[str]:
        return None

    @abstractmethod
    def scope_type(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def scope_id(self) -> str:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def permission_operation_type(cls) -> OperationType:
        raise NotImplementedError


class BaseScopeActionResult(BaseActionResult):
    @override
    def entity_id(self) -> Optional[str]:
        return None

    @abstractmethod
    def scope_type(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def scope_id(self) -> str:
        raise NotImplementedError


TScopeAction = TypeVar("TScopeAction", bound=BaseScopeAction)
TScopeActionResult = TypeVar("TScopeActionResult", bound=BaseScopeActionResult)



"""Deployment execution logic."""

import asyncio
import logging
from collections.abc import Sequence
from typing import Any, Coroutine, Optional
from uuid import UUID

from ai.backend.common.clients.http_client.client_pool import (
    ClientKey,
    ClientPool,
)
from ai.backend.common.clients.valkey_client.valkey_stat.client import ValkeyStatClient
from ai.backend.common.config import ModelHealthCheck
from ai.backend.common.types import (
    RuntimeVariant,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.clients.appproxy.client import AppProxyClient
from ai.backend.manager.clients.appproxy.types import (
    CreateEndpointRequestBody,
    EndpointTagsModel,
    SessionTagsModel,
    TagsModel,
)
from ai.backend.manager.config.provider import ManagerConfigProvider
from ai.backend.manager.data.deployment.types import (
    DeploymentInfo,
    RouteInfo,
    ScaleOutDecision,
)
from ai.backend.manager.data.resource.types import ScalingGroupProxyTarget
from ai.backend.manager.errors.service import ModelDefinitionNotFound
from ai.backend.manager.repositories.deployment.repository import DeploymentRepository
from ai.backend.manager.sokovan.deployment.definition_generator.registry import (
    ModelDefinitionGeneratorRegistry,
    RegistryArgs,
)
from ai.backend.manager.sokovan.scheduling_controller import SchedulingController

from .types import (
    DeploymentExecutionError,
    DeploymentExecutionResult,
)

log = BraceStyleAdapter(logging.getLogger(__name__))

REGISTER_ENDPOINT_TIMEOUT_SEC = 30


class DeploymentExecutor:
    """Executor for deployment operations."""

    _deployment_repo: DeploymentRepository
    _scheduling_controller: SchedulingController
    _config_provider: ManagerConfigProvider
    _client_pool: ClientPool
    _valkey_stat: ValkeyStatClient

    def __init__(
        self,
        deployment_repo: DeploymentRepository,
        scheduling_controller: SchedulingController,
        config_provider: ManagerConfigProvider,
        client_pool: ClientPool,
        valkey_stat: ValkeyStatClient,
    ) -> None:
        """Initialize the deployment executor."""
        self._deployment_repo = deployment_repo
        self._scheduling_controller = scheduling_controller
        self._config_provider = config_provider
        self._client_pool = client_pool
        self._valkey_stat = valkey_stat
        self._model_definition_generator_registry = ModelDefinitionGeneratorRegistry(
            RegistryArgs(deployment_repository=self._deployment_repo)
        )

    async def check_pending_deployments(
        self, deployments: Sequence[DeploymentInfo]
    ) -> DeploymentExecutionResult:
        scaling_groups = set(deployment.metadata.resource_group for deployment in deployments)
        scaling_group_targets = await self._deployment_repo.fetch_scaling_group_proxy_targets(
            scaling_groups
        )

        # Collect registration tasks
        registration_tasks: list[Coroutine[Any, Any, str]] = []
        for deployment in deployments:
            target_revision = deployment.target_revision()
            if not target_revision:
                log.warning(
                    "Deployment {} has no target revision, skipping",
                    deployment.id,
                )
                continue
            targets = scaling_group_targets[deployment.metadata.resource_group]
            if not targets:
                log.warning(
                    "No proxy target found for scaling group {}, skipping deployment {}",
                    deployment.metadata.resource_group,
                    deployment.id,
                )
                continue
            registration_tasks.append(self._register_endpoint(deployment, targets))

        # Wait for all tasks to complete
        successful_deployments = []
        errors: list[DeploymentExecutionError] = []
        url_updates: dict[UUID, str] = {}

        if registration_tasks:
            results = await asyncio.gather(*registration_tasks, return_exceptions=True)
            for deployment, result in zip(deployments, results):
                if isinstance(result, BaseException):
                    log.error(
                        "Failed to register endpoint for deployment {}: {}",
                        deployment.id,
                        result,
                    )
                    errors.append(
                        DeploymentExecutionError(
                            deployment_info=deployment,
                            reason=str(result),
                            error_detail="Failed to register endpoint",
                        )
                    )
                else:
                    # Result is the endpoint URL string returned from _register_endpoint
                    url_updates[deployment.id] = result
                    successful_deployments.append(deployment)
                    log.info(
                        "Successfully registered endpoint for deployment {} with URL: {}",
                        deployment.id,
                        result,
                    )

        # Batch update all endpoint URLs in the database
        if url_updates:
            await self._deployment_repo.update_endpoint_urls_bulk(url_updates)

        return DeploymentExecutionResult(
            successes=successful_deployments,
            errors=errors,
        )

    async def scale_deployment(
        self, deployments: Sequence[DeploymentInfo]
    ) -> DeploymentExecutionResult:
        endpoint_ids = {deployment.id for deployment in deployments}
        route_map = await self._deployment_repo.fetch_active_routes_by_endpoint_ids(endpoint_ids)
        scale_outs: list[ScaleOutDecision] = []
        scale_ins: list[RouteInfo] = []
        successes = []
        errors = []
        for deployment in deployments:
            try:
                target_count = deployment.replica_spec.target_replica_count
                routes = route_map[deployment.id]
                if len(routes) < target_count:
                    scale_outs.append(
                        ScaleOutDecision(
                            deployment_info=deployment,
                            new_replica_count=target_count - len(routes),
                        )
                    )
                elif len(routes) > target_count:
                    termination_route_candidates = sorted(
                        routes, key=lambda r: (r.status.termination_priority())
                    )
                    candidates = termination_route_candidates[: len(routes) - target_count]
                    scale_ins.extend(
                        candidates,
                    )
                successes.append(deployment)
            except Exception as e:
                log.warning("Failed to scale deployment {}: {}", deployment.id, e)
                errors.append(
                    DeploymentExecutionError(
                        deployment_info=deployment,
                        reason=str(e),
                        error_detail="Failed to scale deployment",
                    )
                )
        await self._deployment_repo.scale_routes(scale_outs, scale_ins)
        return DeploymentExecutionResult(
            successes=successes,
            errors=errors,
        )

    async def calculate_desired_replicas(
        self, deployments: Sequence[DeploymentInfo]
    ) -> DeploymentExecutionResult:
        endpoint_ids = {deployment.id for deployment in deployments}
        auto_scaling_rules = await self._deployment_repo.fetch_auto_scaling_rules_by_endpoint_ids(
            endpoint_ids
        )

        # Fetch all metrics data upfront
        metrics_data = await self._deployment_repo.fetch_metrics_for_autoscaling(
            deployments, auto_scaling_rules
        )

        successes: list[DeploymentInfo] = []
        skipped: list[DeploymentInfo] = []
        errors: list[DeploymentExecutionError] = []
        desired_replicas_map: dict[UUID, int] = {}

        for deployment in deployments:
            auto_scaling_rule = auto_scaling_rules.get(deployment.id, [])
            if not auto_scaling_rule:
                routes = metrics_data.routes_by_endpoint.get(deployment.id, [])
                if deployment.replica_spec.replica_count != len(routes):
                    desired_replicas_map[deployment.id] = deployment.replica_spec.replica_count
                    successes.append(deployment)
                else:
                    skipped.append(deployment)
                continue
            try:
                # Calculate desired replicas
                desired_replica = (
                    await self._deployment_repo.calculate_desired_replicas_for_deployment(
                        deployment,
                        auto_scaling_rule,
                        metrics_data,
                    )
                )

                if desired_replica is None:
                    log.debug(
                        "No change in desired replicas for deployment {}, skipping", deployment.id
                    )
                    skipped.append(deployment)
                    continue

                desired_replicas_map[deployment.id] = desired_replica
                successes.append(deployment)
            except Exception as e:
                log.warning(
                    "Failed to calculate desired replicas for deployment {}: {}", deployment.id, e
                )
                errors.append(
                    DeploymentExecutionError(
                        deployment_info=deployment,
                        reason=str(e),
                        error_detail="Failed to calculate desired replicas",
                    )
                )
                continue

        # Batch update desired replicas
        if desired_replicas_map:
            await self._deployment_repo.update_desired_replicas_bulk(desired_replicas_map)

        return DeploymentExecutionResult(
            successes=successes,
            skipped=skipped,
            errors=errors,
        )

    async def destroy_deployment(
        self, deployments: Sequence[DeploymentInfo]
    ) -> DeploymentExecutionResult:
        endpoint_ids = {deployment.id for deployment in deployments}
        scaling_groups = {deployment.metadata.resource_group for deployment in deployments}
        routes = await self._deployment_repo.fetch_active_routes_by_endpoint_ids(endpoint_ids)
        proxy_targets = await self._deployment_repo.fetch_scaling_group_proxy_targets(
            scaling_groups
        )
        successes = []
        errors = []
        route_ids = set()
        for route_list in routes.values():
            for route in route_list:
                route_ids.add(route.route_id)
        await self._deployment_repo.mark_terminating_route_status_bulk(route_ids)
        for deployment in deployments:
            proxy_target = proxy_targets[deployment.metadata.resource_group]
            if not proxy_target:
                log.warning(
                    "No proxy target found for scaling group {}, skipping deployment {}",
                    deployment.metadata.resource_group,
                    deployment.id,
                )
                errors.append(
                    DeploymentExecutionError(
                        deployment_info=deployment,
                        reason="No proxy target found",
                        error_detail="Cannot delete endpoint from WSProxy",
                    )
                )
                continue
            try:
                await self._delete_endpoint_from_wsproxy(
                    deployment.id,
                    proxy_target.addr,
                    proxy_target.api_token,
                )
                successes.append(deployment)
            except Exception:
                log.warning(
                    "Failed to delete endpoint {} from WSProxy, it might have been already removed",
                    deployment.id,
                )
                errors.append(
                    DeploymentExecutionError(
                        deployment_info=deployment,
                        reason="Failed to delete endpoint from WSProxy",
                        error_detail="Endpoint might have been already removed",
                    )
                )
            log.info("Deleted endpoint {} from WSProxy", deployment.id)
        return DeploymentExecutionResult(
            successes=successes,
            errors=errors,
        )

    # Private helper methods

    async def _register_endpoint(
        self,
        deployment: DeploymentInfo,
        scaling_group_target: ScalingGroupProxyTarget,
    ) -> str:
        target_revision = deployment.target_revision()
        if not target_revision:
            raise ModelDefinitionNotFound(f"No target revision for deployment {deployment.id}")
        generator = self._model_definition_generator_registry.get(
            target_revision.execution.runtime_variant
        )
        model_definition = await generator.generate_model_definition(target_revision)
        health_check_config = model_definition.health_check_config()
        if not health_check_config:
            log.debug(
                "No health check configuration found in model definition for deployment {}",
                deployment.id,
            )
        return await self._create_endpoint_in_proxy(
            endpoint_id=deployment.id,
            endpoint_name=deployment.metadata.name,
            session_owner_id=deployment.metadata.session_owner,
            project_id=deployment.metadata.project,
            domain_name=deployment.metadata.domain,
            runtime_variant=target_revision.execution.runtime_variant,
            existing_url=deployment.network.url,
            open_to_public=deployment.network.open_to_public,
            health_check_config=health_check_config,
            app_proxy_addr=scaling_group_target.addr,
            app_proxy_api_token=scaling_group_target.api_token,
        )

    def _load_app_proxy_client(self, address: str, token: str) -> AppProxyClient:
        """Load or create a App Proxy client for the given address."""
        client_session = self._client_pool.load_client_session(
            ClientKey(
                endpoint=address,
                domain="wsproxy",
            )
        )
        return AppProxyClient(client_session, address, token)

    async def _create_endpoint_in_proxy(
        self,
        endpoint_id: UUID,
        endpoint_name: str,
        session_owner_id: UUID,
        project_id: UUID,
        domain_name: str,
        runtime_variant: RuntimeVariant,
        existing_url: Optional[str],
        open_to_public: bool,
        health_check_config: Optional[ModelHealthCheck],
        app_proxy_addr: str,
        app_proxy_api_token: str,
    ) -> str:
        """
        Create an endpoint in WSProxy service.

        Args:
            endpoint_id: Endpoint UUID
            endpoint_name: Name of the endpoint
            session_owner_id: UUID of the session owner
            project_id: UUID of the project
            domain_name: Domain name
            runtime_variant: Runtime variant
            existing_url: Existing URL if any
            open_to_public: Public accessibility flag
            health_check_config: Optional health check configuration
            wsproxy_addr: WSProxy service address
            wsproxy_api_token: WSProxy API token

        Returns:
            Response from WSProxy service
        """
        app_proxy_client = self._load_app_proxy_client(app_proxy_addr, app_proxy_api_token)

        # Create request body using Pydantic model
        request_body = CreateEndpointRequestBody(
            version="v2",
            service_name=endpoint_name,
            tags=TagsModel(
                session=SessionTagsModel(
                    user_uuid=str(session_owner_id),
                    group_id=str(project_id),
                    domain_name=domain_name,
                ),
                endpoint=EndpointTagsModel(
                    id=str(endpoint_id),
                    runtime_variant=runtime_variant.value,
                    existing_url=str(existing_url) if existing_url else None,
                ),
            ),
            apps={},
            open_to_public=open_to_public,
            health_check=health_check_config,
        )

        res = await app_proxy_client.create_endpoint(endpoint_id, request_body)
        return res["endpoint"]

    async def _delete_endpoint_from_wsproxy(
        self,
        endpoint_id: UUID,
        app_proxy_addr: str,
        app_proxy_api_token: str,
    ) -> None:
        """
        Delete an endpoint from WSProxy service.

        Args:
            endpoint_id: Endpoint UUID to delete
            wsproxy_addr: WSProxy service address
            wsproxy_api_token: WSProxy API token
        """
        app_proxy_client = self._load_app_proxy_client(app_proxy_addr, app_proxy_api_token)
        await app_proxy_client.delete_endpoint(endpoint_id)



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.model_serving.creator import ModelServiceCreator
from ai.backend.manager.data.model_serving.types import ServiceInfo
from ai.backend.manager.services.model_serving.actions.base import ModelServiceAction


@dataclass
class CreateModelServiceAction(ModelServiceAction):
    request_user_id: uuid.UUID
    creator: ModelServiceCreator

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "create"


@dataclass
class CreateModelServiceActionResult(BaseActionResult):
    data: ServiceInfo

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.data.endpoint_id)



"""Processors for app configuration service."""

from __future__ import annotations

from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec

from .actions import (
    DeleteDomainConfigAction,
    DeleteDomainConfigActionResult,
    DeleteUserConfigAction,
    DeleteUserConfigActionResult,
    GetDomainConfigAction,
    GetDomainConfigActionResult,
    GetMergedAppConfigAction,
    GetMergedAppConfigActionResult,
    GetUserConfigAction,
    GetUserConfigActionResult,
    UpsertDomainConfigAction,
    UpsertDomainConfigActionResult,
    UpsertUserConfigAction,
    UpsertUserConfigActionResult,
)
from .service import AppConfigService


class AppConfigProcessors(AbstractProcessorPackage):
    """Processors for app configuration operations."""

    # Domain config processors
    get_domain_config: ActionProcessor[GetDomainConfigAction, GetDomainConfigActionResult]
    upsert_domain_config: ActionProcessor[UpsertDomainConfigAction, UpsertDomainConfigActionResult]
    delete_domain_config: ActionProcessor[DeleteDomainConfigAction, DeleteDomainConfigActionResult]

    # User config processors
    get_user_config: ActionProcessor[GetUserConfigAction, GetUserConfigActionResult]
    upsert_user_config: ActionProcessor[UpsertUserConfigAction, UpsertUserConfigActionResult]
    delete_user_config: ActionProcessor[DeleteUserConfigAction, DeleteUserConfigActionResult]

    # Merged config processor
    get_merged_config: ActionProcessor[GetMergedAppConfigAction, GetMergedAppConfigActionResult]

    def __init__(self, service: AppConfigService, action_monitors: list[ActionMonitor]) -> None:
        # Domain config processors
        self.get_domain_config = ActionProcessor(service.get_domain_config, action_monitors)
        self.upsert_domain_config = ActionProcessor(service.upsert_domain_config, action_monitors)
        self.delete_domain_config = ActionProcessor(service.delete_domain_config, action_monitors)

        # User config processors
        self.get_user_config = ActionProcessor(service.get_user_config, action_monitors)
        self.upsert_user_config = ActionProcessor(service.upsert_user_config, action_monitors)
        self.delete_user_config = ActionProcessor(service.delete_user_config, action_monitors)

        # Merged config processor
        self.get_merged_config = ActionProcessor(service.get_merged_config, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            # Domain config actions
            GetDomainConfigAction.spec(),
            UpsertDomainConfigAction.spec(),
            DeleteDomainConfigAction.spec(),
            # User config actions
            GetUserConfigAction.spec(),
            UpsertUserConfigAction.spec(),
            DeleteUserConfigAction.spec(),
            # Merged config action
            GetMergedAppConfigAction.spec(),
        ]



"""
Data types for domain service.
Deprecated: use `ai.backend.manager.data.domain.types` instead.
"""

import uuid
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, Optional, Self, override

from sqlalchemy.engine.result import Row

from ai.backend.common.types import ResourceSlot, VFolderHostPermissionMap
from ai.backend.manager.models.domain import DomainRow
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.types import Creator, OptionalState, PartialModifier, TriState


@dataclass
class UserInfo:
    id: uuid.UUID
    role: UserRole
    domain_name: str


@dataclass
class DomainData:
    name: str
    description: Optional[str]
    is_active: bool
    created_at: datetime = field(compare=False)
    modified_at: datetime = field(compare=False)
    total_resource_slots: ResourceSlot
    allowed_vfolder_hosts: VFolderHostPermissionMap
    allowed_docker_registries: list[str]
    dotfiles: bytes
    integration_id: Optional[str]

    @classmethod
    def from_row(cls, row: Optional[DomainRow | Row]) -> Optional[Self]:
        if row is None:
            return None
        return cls(
            name=row.name,
            description=row.description,
            is_active=row.is_active,
            created_at=row.created_at,
            modified_at=row.modified_at,
            total_resource_slots=row.total_resource_slots,
            allowed_vfolder_hosts=row.allowed_vfolder_hosts,
            allowed_docker_registries=row.allowed_docker_registries,
            dotfiles=row.dotfiles,
            integration_id=row.integration_id,
        )


@dataclass
class DomainCreator(Creator):
    name: str
    description: Optional[str] = None
    is_active: Optional[bool] = None
    total_resource_slots: Optional[ResourceSlot] = None
    allowed_vfolder_hosts: Optional[dict[str, list[str]]] = None
    allowed_docker_registries: Optional[list[str]] = None
    integration_id: Optional[str] = None
    dotfiles: Optional[bytes] = None

    @override
    def fields_to_store(self) -> dict[str, Any]:
        to_store: dict[str, Any] = {"name": self.name}
        if self.description is not None:
            to_store["description"] = self.description
        if self.is_active is not None:
            to_store["is_active"] = self.is_active
        if self.total_resource_slots is not None:
            to_store["total_resource_slots"] = self.total_resource_slots
        if self.allowed_vfolder_hosts is not None:
            to_store["allowed_vfolder_hosts"] = self.allowed_vfolder_hosts
        if self.allowed_docker_registries is not None:
            to_store["allowed_docker_registries"] = self.allowed_docker_registries
        if self.integration_id is not None:
            to_store["integration_id"] = self.integration_id
        if self.dotfiles is not None:
            to_store["dotfiles"] = self.dotfiles
        return to_store


@dataclass
class DomainModifier(PartialModifier):
    name: OptionalState[str] = field(default_factory=OptionalState.nop)
    description: TriState[str] = field(default_factory=TriState.nop)
    is_active: OptionalState[bool] = field(default_factory=OptionalState.nop)
    total_resource_slots: TriState[ResourceSlot] = field(default_factory=TriState.nop)
    allowed_vfolder_hosts: OptionalState[dict[str, list[str]]] = field(
        default_factory=OptionalState.nop
    )
    allowed_docker_registries: OptionalState[list[str]] = field(default_factory=OptionalState.nop)
    integration_id: TriState[str] = field(default_factory=TriState.nop)

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.description.update_dict(to_update, "description")
        self.is_active.update_dict(to_update, "is_active")
        self.total_resource_slots.update_dict(to_update, "total_resource_slots")
        self.allowed_vfolder_hosts.update_dict(to_update, "allowed_vfolder_hosts")
        self.allowed_docker_registries.update_dict(to_update, "allowed_docker_registries")
        self.integration_id.update_dict(to_update, "integration_id")
        return to_update


@dataclass
class DomainNodeModifier(PartialModifier):
    description: TriState[str] = field(default_factory=TriState[str].nop)
    is_active: OptionalState[bool] = field(default_factory=OptionalState[bool].nop)
    total_resource_slots: TriState[ResourceSlot] = field(default_factory=TriState[ResourceSlot].nop)
    allowed_vfolder_hosts: OptionalState[dict[str, list[str]]] = field(
        default_factory=OptionalState[dict[str, list[str]]].nop
    )
    allowed_docker_registries: OptionalState[list[str]] = field(
        default_factory=OptionalState[list[str]].nop
    )
    integration_id: TriState[str] = field(default_factory=TriState[str].nop)
    dotfiles: OptionalState[bytes] = field(default_factory=OptionalState[bytes].nop)

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.description.update_dict(to_update, "description")
        self.is_active.update_dict(to_update, "is_active")
        self.total_resource_slots.update_dict(to_update, "total_resource_slots")
        self.allowed_vfolder_hosts.update_dict(to_update, "allowed_vfolder_hosts")
        self.allowed_docker_registries.update_dict(to_update, "allowed_docker_registries")
        self.integration_id.update_dict(to_update, "integration_id")
        self.dotfiles.update_dict(to_update, "dotfiles")
        return to_update



from datetime import datetime
from typing import Self
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.manager.api.gql.base import OrderDirection
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.data.deployment.access_token import ModelDeploymentAccessTokenCreator
from ai.backend.manager.data.deployment.types import (
    AccessTokenOrderField,
    ModelDeploymentAccessTokenData,
)
from ai.backend.manager.errors.common import ServerMisconfiguredError
from ai.backend.manager.services.deployment.actions.access_token.create_access_token import (
    CreateAccessTokenAction,
)


@strawberry.input(description="Added in 25.16.0")
class AccessTokenOrderBy:
    field: AccessTokenOrderField
    direction: OrderDirection = OrderDirection.DESC


@strawberry.type
class AccessToken(Node):
    id: NodeID[str]
    token: str = strawberry.field(description="Added in 25.16.0: The access token.")
    created_at: datetime = strawberry.field(
        description="Added in 25.16.0: The creation timestamp of the access token."
    )
    valid_until: datetime = strawberry.field(
        description="Added in 25.16.0: The expiration timestamp of the access token."
    )

    @classmethod
    def from_dataclass(cls, data: ModelDeploymentAccessTokenData) -> Self:
        return cls(
            id=ID(str(data.id)),
            token=data.token,
            created_at=data.created_at,
            valid_until=data.valid_until,
        )


AccessTokenEdge = Edge[AccessToken]


@strawberry.type(description="Added in 25.16.0")
class AccessTokenConnection(Connection[AccessToken]):
    count: int

    def __init__(self, *args, count: int, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = count


@strawberry.input
class CreateAccessTokenInput:
    model_deployment_id: ID = strawberry.field(
        description="Added in 25.16.0: The ID of the model deployment for which the access token is created."
    )
    valid_until: datetime = strawberry.field(
        description="Added in 25.16.0: The expiration timestamp of the access token."
    )

    def to_creator(self) -> "ModelDeploymentAccessTokenCreator":
        return ModelDeploymentAccessTokenCreator(
            model_deployment_id=UUID(self.model_deployment_id),
            valid_until=self.valid_until,
        )


@strawberry.type
class CreateAccessTokenPayload:
    access_token: AccessToken


@strawberry.mutation(description="Added in 25.16.0")
async def create_access_token(
    input: CreateAccessTokenInput, info: Info[StrawberryGQLContext]
) -> CreateAccessTokenPayload:
    deployment_processor = info.context.processors.deployment
    if deployment_processor is None:
        raise ServerMisconfiguredError("Deployment processor is not available")
    result = await deployment_processor.create_access_token.wait_for_complete(
        action=CreateAccessTokenAction(input.to_creator())
    )
    return CreateAccessTokenPayload(access_token=AccessToken.from_dataclass(result.data))



import uuid

from ai.backend.manager.data.model_serving.types import RequesterCtx
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.services.model_serving.exceptions import (
    GenericForbidden,
    InvalidAPIParameters,
)
from ai.backend.manager.utils import check_if_requester_is_eligible_to_act_as_target_user_uuid


async def verify_user_access_scopes(
    db: ExtendedAsyncSAEngine, requester_ctx: RequesterCtx, owner_uuid: uuid.UUID
) -> None:
    if requester_ctx.is_authorized is False:
        raise GenericForbidden("Only authorized requests may have access key scopes.")
    if owner_uuid is None or owner_uuid == requester_ctx.user_id:
        return
    async with db.begin_readonly() as conn:
        try:
            await check_if_requester_is_eligible_to_act_as_target_user_uuid(
                conn,
                requester_ctx.user_role,
                requester_ctx.domain_name,
                owner_uuid,
            )
            return
        except ValueError as e:
            raise InvalidAPIParameters(str(e))
        except RuntimeError as e:
            raise GenericForbidden(str(e))



from __future__ import annotations

from typing import TYPE_CHECKING

from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience import (
    MetricArgs,
    MetricPolicy,
    Resilience,
    RetryArgs,
    RetryPolicy,
)
from ai.backend.common.resilience.policies.retry import BackoffStrategy
from ai.backend.manager.data.scaling_group.types import ScalingGroupListResult
from ai.backend.manager.repositories.base import Querier

from .db_source import ScalingGroupDBSource

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("ScalingGroupRepository",)


scaling_group_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.SCALING_GROUP_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
            )
        ),
    ]
)


class ScalingGroupRepository:
    """Repository for scaling group-related data access."""

    _db_source: ScalingGroupDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = ScalingGroupDBSource(db)

    @scaling_group_repository_resilience.apply()
    async def search_scaling_groups(
        self,
        querier: Querier,
    ) -> ScalingGroupListResult:
        """Searches scaling groups with total count."""
        return await self._db_source.search_scaling_groups(querier=querier)



from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import Any


class AbstractConfigLoader(ABC):
    @abstractmethod
    async def load(self) -> Mapping[str, Any]:
        raise NotImplementedError



"""Exceptions for validators."""

from aiohttp import web

from ai.backend.common.exception import (
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.manager.sokovan.scheduler.exceptions import SchedulingError
from ai.backend.manager.sokovan.scheduler.types import SchedulingPredicate


class SchedulingValidationError(SchedulingError, web.HTTPPreconditionFailed):
    """Base exception for validation errors in the Sokovan scheduler."""

    error_type = "https://api.backend.ai/probs/scheduling-validation-failed"
    error_title = "Scheduling validation failed."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.SCHEDULE,
            error_detail=ErrorDetail.FORBIDDEN,
        )

    def failed_predicates(self) -> list[SchedulingPredicate]:
        """Return list of failed predicates for this error."""
        return [SchedulingPredicate(name=type(self).__name__, msg=str(self))]


class ConcurrencyLimitExceeded(SchedulingValidationError):
    """Raised when concurrent session limit is exceeded."""

    error_type = "https://api.backend.ai/probs/concurrency-limit-exceeded"
    error_title = "Concurrent session limit exceeded."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.CHECK_LIMIT,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class DependenciesNotSatisfied(SchedulingValidationError):
    """Raised when session dependencies are not satisfied."""

    error_type = "https://api.backend.ai/probs/dependencies-not-satisfied"
    error_title = "Session dependencies not satisfied."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.CHECK_LIMIT,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class KeypairResourceQuotaExceeded(SchedulingValidationError):
    """Raised when keypair resource quota is exceeded."""

    error_type = "https://api.backend.ai/probs/keypair-resource-quota-exceeded"
    error_title = "Keypair resource quota exceeded."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KEYPAIR,
            operation=ErrorOperation.CHECK_LIMIT,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class UserResourceQuotaExceeded(SchedulingValidationError):
    """Raised when user resource quota is exceeded."""

    error_type = "https://api.backend.ai/probs/user-resource-quota-exceeded"
    error_title = "User resource quota exceeded."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.USER,
            operation=ErrorOperation.CHECK_LIMIT,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class GroupResourceQuotaExceeded(SchedulingValidationError):
    """Raised when group resource quota is exceeded."""

    error_type = "https://api.backend.ai/probs/group-resource-quota-exceeded"
    error_title = "Group resource quota exceeded."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.GROUP,
            operation=ErrorOperation.CHECK_LIMIT,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class DomainResourceQuotaExceeded(SchedulingValidationError):
    """Raised when domain resource quota is exceeded."""

    error_type = "https://api.backend.ai/probs/domain-resource-quota-exceeded"
    error_title = "Domain resource quota exceeded."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.DOMAIN,
            operation=ErrorOperation.CHECK_LIMIT,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class PendingSessionCountLimitExceeded(SchedulingValidationError):
    """Raised when pending session count limit is exceeded."""

    error_type = "https://api.backend.ai/probs/pending-session-count-limit-exceeded"
    error_title = "Pending session count limit exceeded."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.CHECK_LIMIT,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class PendingSessionResourceLimitExceeded(SchedulingValidationError):
    """Raised when pending session resource limit is exceeded."""

    error_type = "https://api.backend.ai/probs/pending-session-resource-limit-exceeded"
    error_title = "Pending session resource limit exceeded."

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.SESSION,
            operation=ErrorOperation.CHECK_LIMIT,
            error_detail=ErrorDetail.FORBIDDEN,
        )


class MultipleValidationErrors(SchedulingValidationError):
    """Raised when multiple validation errors occur."""

    error_type = "https://api.backend.ai/probs/multiple-validation-errors"
    error_title = "Multiple validation errors occurred."

    _errors: list[SchedulingValidationError]

    def __init__(self, errors: list[SchedulingValidationError]) -> None:
        """
        Initialize with a list of validation errors.

        Args:
            errors: List of validation errors that occurred
        """
        self._errors = errors
        # Format each error on a new line for better readability
        messages = []
        for i, e in enumerate(errors, 1):
            error_name = type(e).__name__
            error_msg = str(e)
            messages.append(f"  {i}. {error_name}: {error_msg}")

        # Join with newlines for better formatting
        formatted_errors = "\n".join(messages)
        super().__init__(f"Multiple validation errors occurred:\n{formatted_errors}")

    def failed_predicates(self) -> list[SchedulingPredicate]:
        """Return list of all failed predicates from all errors."""
        predicates = []
        for error in self._errors:
            predicates.extend(error.failed_predicates())
        return predicates

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.BACKENDAI,
            operation=ErrorOperation.SCHEDULE,
            error_detail=ErrorDetail.FORBIDDEN,
        )



from .composer import InfrastructureComposer, InfrastructureInput, InfrastructureResources
from .redis import ValkeyClients

__all__ = [
    "ValkeyClients",
    "InfrastructureInput",
    "InfrastructureResources",
    "InfrastructureComposer",
]



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.deployment.types import ModelRevisionData
from ai.backend.manager.repositories.deployment.types.types import (
    ModelRevisionFilterOptions,
    ModelRevisionOrderingOptions,
)
from ai.backend.manager.services.deployment.actions.model_revision.base import (
    ModelRevisionBaseAction,
)
from ai.backend.manager.types import PaginationOptions


@dataclass
class ListRevisionsAction(ModelRevisionBaseAction):
    pagination: PaginationOptions
    ordering: Optional[ModelRevisionOrderingOptions] = None
    filters: Optional[ModelRevisionFilterOptions] = None

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "list_revisions"


@dataclass
class ListRevisionsActionResult(BaseActionResult):
    data: list[ModelRevisionData]
    total_count: int

    @override
    def entity_id(self) -> Optional[str]:
        return None



"""Database source for notification repository operations."""

from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.orm import selectinload

from ai.backend.manager.data.notification import (
    NotificationChannelCreator,
    NotificationChannelData,
    NotificationChannelListResult,
    NotificationChannelModifier,
    NotificationRuleCreator,
    NotificationRuleData,
    NotificationRuleListResult,
    NotificationRuleModifier,
    NotificationRuleType,
)
from ai.backend.manager.errors.notification import (
    NotificationChannelNotFound,
    NotificationRuleNotFound,
)
from ai.backend.manager.models.notification import (
    NotificationChannelRow,
    NotificationRuleRow,
)
from ai.backend.manager.repositories.base import Querier, execute_querier

if TYPE_CHECKING:
    from sqlalchemy.ext.asyncio import AsyncSession as SASession

    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine


__all__ = (
    "NotificationDBSource",
    "NotificationChannelListResult",
    "NotificationRuleListResult",
)


class NotificationDBSource:
    """
    Database source for notification operations.
    Handles all database operations for notification channels and rules.
    """

    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def get_matching_rules(
        self,
        rule_type: NotificationRuleType,
        enabled_only: bool = True,
    ) -> list[NotificationRuleData]:
        """Retrieves all notification rules that match the given rule type."""
        async with self._db.begin_readonly_session() as db_sess:
            rows = await self._fetch_matching_rules(db_sess, rule_type, enabled_only)
            return [row.to_data() for row in rows]

    async def _fetch_matching_rules(
        self,
        db_sess: SASession,
        rule_type: NotificationRuleType,
        enabled_only: bool,
    ) -> list[NotificationRuleRow]:
        """Private method to fetch matching rules with channel loaded."""
        query = (
            sa.select(NotificationRuleRow)
            .where(NotificationRuleRow.rule_type == str(rule_type))
            .options(selectinload(NotificationRuleRow.channel))
        )

        if enabled_only:
            query = query.where(NotificationRuleRow.enabled == sa.true())

        result = await db_sess.execute(query)
        rows = list(result.scalars().all())

        # Filter by channel enabled status if needed
        if enabled_only:
            rows = [row for row in rows if row.channel.enabled]

        return rows

    async def create_channel(
        self,
        creator: NotificationChannelCreator,
    ) -> NotificationChannelData:
        """Creates a new notification channel."""
        async with self._db.begin_session() as db_sess:
            channel_row = NotificationChannelRow(**creator.fields_to_store())
            db_sess.add(channel_row)
            await db_sess.flush()
            await db_sess.refresh(channel_row)
            return channel_row.to_data()

    async def update_channel(
        self,
        channel_id: UUID,
        modifier: NotificationChannelModifier,
    ) -> NotificationChannelData:
        """Updates an existing notification channel."""
        async with self._db.begin_session() as db_sess:
            values = modifier.fields_to_update()
            if not values:
                raise NotificationChannelNotFound("No fields to update for notification channel")

            stmt = (
                sa.update(NotificationChannelRow)
                .where(NotificationChannelRow.id == channel_id)
                .values(values)
            )
            result = await db_sess.execute(stmt)
            if result.rowcount == 0:
                raise NotificationChannelNotFound(f"Notification channel {channel_id} not found")

            # Fetch the updated row
            select_stmt = sa.select(NotificationChannelRow).where(
                NotificationChannelRow.id == channel_id
            )
            result = await db_sess.execute(select_stmt)
            row = result.scalar_one()
            return row.to_data()

    async def delete_channel(self, channel_id: UUID) -> bool:
        """Deletes a notification channel."""
        async with self._db.begin_session() as db_sess:
            stmt = sa.delete(NotificationChannelRow).where(NotificationChannelRow.id == channel_id)
            result = await db_sess.execute(stmt)
            return result.rowcount > 0

    async def create_rule(
        self,
        creator: NotificationRuleCreator,
    ) -> NotificationRuleData:
        """Creates a new notification rule."""
        async with self._db.begin_session() as db_sess:
            rule_row = NotificationRuleRow(**creator.fields_to_store())
            db_sess.add(rule_row)
            await db_sess.flush()
            # Explicitly load the channel relationship
            await db_sess.refresh(rule_row)
            stmt = (
                sa.select(NotificationRuleRow)
                .where(NotificationRuleRow.id == rule_row.id)
                .options(selectinload(NotificationRuleRow.channel))
            )
            result = await db_sess.execute(stmt)
            row = result.scalar_one()
            return row.to_data()

    async def update_rule(
        self,
        rule_id: UUID,
        modifier: NotificationRuleModifier,
    ) -> NotificationRuleData:
        """Updates an existing notification rule."""
        async with self._db.begin_session() as db_sess:
            values = modifier.fields_to_update()
            if not values:
                raise NotificationRuleNotFound("No fields to update for notification rule")

            stmt = (
                sa.update(NotificationRuleRow)
                .where(NotificationRuleRow.id == rule_id)
                .values(values)
            )
            result = await db_sess.execute(stmt)
            if result.rowcount == 0:
                raise NotificationRuleNotFound(f"Notification rule {rule_id} not found")

            # Fetch the updated row with channel relationship loaded
            select_stmt = (
                sa.select(NotificationRuleRow)
                .where(NotificationRuleRow.id == rule_id)
                .options(selectinload(NotificationRuleRow.channel))
            )
            result = await db_sess.execute(select_stmt)
            row = result.scalar_one()
            return row.to_data()

    async def delete_rule(self, rule_id: UUID) -> bool:
        """Deletes a notification rule."""
        async with self._db.begin_session() as db_sess:
            stmt = sa.delete(NotificationRuleRow).where(NotificationRuleRow.id == rule_id)
            result = await db_sess.execute(stmt)
            return result.rowcount > 0

    async def get_channel_by_id(self, channel_id: UUID) -> NotificationChannelData:
        """Retrieves a notification channel by ID."""
        async with self._db.begin_readonly_session() as db_sess:
            row = await db_sess.get(NotificationChannelRow, channel_id)
            if not row:
                raise NotificationChannelNotFound(f"Notification channel {channel_id} not found")
            return row.to_data()

    async def get_rule_by_id(self, rule_id: UUID) -> NotificationRuleData:
        """Retrieves a notification rule by ID."""
        async with self._db.begin_readonly_session() as db_sess:
            stmt = (
                sa.select(NotificationRuleRow)
                .where(NotificationRuleRow.id == rule_id)
                .options(selectinload(NotificationRuleRow.channel))
            )
            result = await db_sess.execute(stmt)
            row = result.scalar_one_or_none()
            if not row:
                raise NotificationRuleNotFound(f"Notification rule {rule_id} not found")
            return row.to_data()

    async def search_channels(
        self,
        querier: Querier,
    ) -> NotificationChannelListResult:
        """Searches notification channels with total count."""
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(NotificationChannelRow)

            result = await execute_querier(
                db_sess,
                query,
                querier,
            )

            items = [row.NotificationChannelRow.to_data() for row in result.rows]

            return NotificationChannelListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )

    async def search_rules(
        self,
        querier: Querier,
    ) -> NotificationRuleListResult:
        """Searches notification rules with total count."""
        async with self._db.begin_readonly_session() as db_sess:
            query = sa.select(NotificationRuleRow).options(
                selectinload(NotificationRuleRow.channel)
            )

            result = await execute_querier(
                db_sess,
                query,
                querier,
            )

            items = [row.NotificationRuleRow.to_data() for row in result.rows]

            return NotificationRuleListResult(
                items=items,
                total_count=result.total_count,
                has_next_page=result.has_next_page,
                has_previous_page=result.has_previous_page,
            )



"""Base GraphQL adapter providing common utilities."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Optional, final

from ai.backend.manager.api.gql.base import decode_cursor
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy
from ai.backend.manager.errors.api import InvalidGraphQLParameters
from ai.backend.manager.repositories.base import (
    CursorBackwardPagination,
    CursorConditionFactory,
    CursorForwardPagination,
    OffsetPagination,
    Querier,
    QueryCondition,
    QueryOrder,
    QueryPagination,
)

DEFAULT_PAGINATION_LIMIT = 10


@dataclass(frozen=True)
class PaginationOptions:
    """GraphQL pagination arguments."""

    first: Optional[int] = None
    after: Optional[str] = None
    last: Optional[int] = None
    before: Optional[str] = None
    limit: Optional[int] = None
    offset: Optional[int] = None


@dataclass(frozen=True)
class PaginationSpec:
    """Specification for pagination behavior.

    Contains domain-specific configuration for pagination:
    - Forward/backward orders and condition factories for cursor-based pagination
    - forward_order is also used as default order for offset pagination when order_by is not provided

    For typical "newest first" lists:
    - Forward (first/after): DESC order, shows newer items first, next page shows older items
    - Backward (last/before): ASC order, fetches older items first (reversed for display)
    """

    forward_order: QueryOrder
    """Order for forward pagination (e.g., created_at DESC for newest first).
    Also used as default order for offset pagination when order_by is not provided."""

    backward_order: QueryOrder
    """Order for backward pagination (e.g., created_at ASC, results reversed for display)."""

    forward_condition_factory: CursorConditionFactory
    """Factory that creates cursor condition for forward pagination (e.g., created_at < cursor)."""

    backward_condition_factory: CursorConditionFactory
    """Factory that creates cursor condition for backward pagination (e.g., created_at > cursor)."""


class BaseGQLAdapter:
    """Base adapter providing common GraphQL query building utilities."""

    def _build_pagination(
        self,
        options: PaginationOptions,
        spec: PaginationSpec,
    ) -> QueryPagination:
        """Build pagination from GraphQL pagination arguments.

        For cursor-based pagination (first/after or last/before), condition factories
        and orders are used from the spec. The factories create cursor conditions from
        decoded cursor values.

        If no pagination parameters are provided, returns a default OffsetPagination
        with limit=DEFAULT_PAGINATION_LIMIT.
        """
        # Validate and build pagination
        # Count how many pagination modes are being used
        pagination_modes = sum([
            options.first is not None,
            options.last is not None,
            options.limit is not None,
        ])

        if pagination_modes > 1:
            raise InvalidGraphQLParameters(
                "Only one pagination mode allowed: (first/after) OR (last/before) OR (limit/offset)"
            )

        # Build appropriate pagination based on parameters
        if options.first is not None:
            if options.first <= 0:
                raise InvalidGraphQLParameters(f"first must be positive, got {options.first}")
            cursor_condition = None
            if options.after is not None:
                cursor_value = decode_cursor(options.after)
                cursor_condition = spec.forward_condition_factory(cursor_value)
            return CursorForwardPagination(
                first=options.first,
                cursor_order=spec.forward_order,
                cursor_condition=cursor_condition,
            )
        elif options.last is not None:
            if options.last <= 0:
                raise InvalidGraphQLParameters(f"last must be positive, got {options.last}")
            cursor_condition = None
            if options.before is not None:
                cursor_value = decode_cursor(options.before)
                cursor_condition = spec.backward_condition_factory(cursor_value)
            return CursorBackwardPagination(
                last=options.last,
                cursor_order=spec.backward_order,
                cursor_condition=cursor_condition,
            )
        elif options.limit is not None:
            if options.limit <= 0:
                raise InvalidGraphQLParameters(f"limit must be positive, got {options.limit}")
            if options.offset is not None and options.offset < 0:
                raise InvalidGraphQLParameters(f"offset must be non-negative, got {options.offset}")
            return OffsetPagination(limit=options.limit, offset=options.offset or 0)

        # Default pagination when no parameters provided
        return OffsetPagination(limit=DEFAULT_PAGINATION_LIMIT, offset=0)

    @final
    def build_querier(
        self,
        pagination_options: PaginationOptions,
        pagination_spec: PaginationSpec,
        filter: Optional[GQLFilter] = None,
        order_by: Optional[Sequence[GQLOrderBy]] = None,
    ) -> Querier:
        """Build Querier from GraphQL arguments with domain configuration.

        Args:
            pagination_options: Pagination parameters (first/after/last/before/limit/offset)
            pagination_spec: Domain-specific pagination specification (orders, condition factories)
            filter: Optional filter with build_conditions() method
            order_by: Optional sequence of order specifications with to_query_order() method

        Returns:
            A Querier instance with conditions, orders, and pagination configured.

        Raises:
            InvalidGraphQLParameters: If order_by is used with cursor pagination.
        """
        # Cursor pagination and order_by are mutually exclusive
        is_cursor_pagination = (
            pagination_options.first is not None or pagination_options.last is not None
        )
        if is_cursor_pagination and order_by is not None:
            raise InvalidGraphQLParameters(
                "order_by cannot be used with cursor pagination (first/after, last/before)"
            )

        conditions: list[QueryCondition] = []
        orders: list[QueryOrder] = []

        if filter:
            conditions.extend(filter.build_conditions())

        if order_by:
            for o in order_by:
                orders.append(o.to_query_order())
        elif not is_cursor_pagination:
            # Apply default order for offset pagination when order_by is not provided
            orders.append(pagination_spec.forward_order)

        pagination = self._build_pagination(
            options=pagination_options,
            spec=pagination_spec,
        )

        return Querier(conditions=conditions, orders=orders, pagination=pagination)



"""Type definitions for WSProxy client."""

from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, ConfigDict, Field

from ai.backend.common.config import ModelHealthCheck


class SessionTagsModel(BaseModel):
    """Session information for endpoint creation."""

    model_config = ConfigDict(populate_by_name=True)

    user_uuid: str
    group_id: str
    domain_name: str


class EndpointTagsModel(BaseModel):
    """Endpoint metadata for endpoint creation."""

    model_config = ConfigDict(populate_by_name=True)

    id: str
    runtime_variant: str
    existing_url: Optional[str] = None


class TagsModel(BaseModel):
    """Combined tags for endpoint creation."""

    model_config = ConfigDict(populate_by_name=True)

    session: SessionTagsModel
    endpoint: EndpointTagsModel


class CreateEndpointRequestBody(BaseModel):
    """Request body for creating an endpoint in WSProxy."""

    model_config = ConfigDict(populate_by_name=True)

    version: str = Field(default="v2", description="API version")
    service_name: str = Field(description="Name of the service/endpoint")
    tags: TagsModel = Field(description="Metadata tags for the endpoint")
    apps: dict = Field(default_factory=dict, description="Application configuration")
    open_to_public: bool = Field(
        default=False, description="Whether the endpoint is publicly accessible"
    )
    health_check: Optional[ModelHealthCheck] = Field(
        default=None, description="Health check configuration"
    )



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Optional, Self

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from ai.backend.manager.data.permission.id import ObjectId
from ai.backend.manager.data.permission.object_permission import (
    ObjectPermissionCreateInput,
    ObjectPermissionData,
)
from ai.backend.manager.data.permission.types import (
    EntityType,
    OperationType,
)

from ...base import (
    GUID,
    Base,
    IDColumn,
    StrEnumType,
)

if TYPE_CHECKING:
    from ..role import RoleRow


class ObjectPermissionRow(Base):
    __tablename__ = "object_permissions"
    __table_args__ = (sa.Index("ix_id_role_id_entity_id", "id", "role_id", "entity_id"),)

    id: uuid.UUID = IDColumn()
    role_id: uuid.UUID = sa.Column("role_id", GUID, nullable=False)
    entity_type: EntityType = sa.Column(
        "entity_type", StrEnumType(EntityType, length=32), nullable=False
    )
    entity_id: str = sa.Column(
        "entity_id", sa.String(64), nullable=False
    )  # e.g., "project_id", "user_id" etc.
    operation: OperationType = sa.Column(
        "operation", StrEnumType(OperationType, length=32), nullable=False
    )

    role_row: Optional[RoleRow] = relationship(
        "RoleRow",
        back_populates="object_permission_rows",
        primaryjoin="RoleRow.id == foreign(ObjectPermissionRow.role_id)",
    )

    def object_id(self) -> ObjectId:
        return ObjectId(entity_type=self.entity_type, entity_id=self.entity_id)

    @classmethod
    def from_input(cls, input: ObjectPermissionCreateInput) -> Self:
        return cls(
            role_id=input.role_id,
            entity_type=input.entity_type,
            entity_id=input.entity_id,
            operation=input.operation,
        )

    @classmethod
    def from_sa_row(cls, row: sa.engine.Row) -> Self:
        return cls(
            id=row.id,
            role_id=row.role_id,
            entity_type=row.entity_type,
            entity_id=row.entity_id,
            operation=row.operation,
        )

    def to_data(self) -> ObjectPermissionData:
        return ObjectPermissionData(
            id=self.id,
            role_id=self.role_id,
            object_id=self.object_id(),
            operation=self.operation,
        )



from collections.abc import Sequence
from datetime import datetime
from typing import AsyncGenerator, Optional
from uuid import UUID, uuid4

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.common.contexts.user import current_user
from ai.backend.common.data.model_deployment.types import (
    DeploymentStrategy as CommonDeploymentStrategy,
)
from ai.backend.common.data.model_deployment.types import (
    ModelDeploymentStatus as CommonDeploymentStatus,
)
from ai.backend.common.exception import ModelDeploymentNotFound, ModelDeploymentUnavailable
from ai.backend.manager.api.gql.base import (
    OrderDirection,
    StringFilter,
    build_page_info,
    build_pagination_options,
    resolve_global_id,
    to_global_id,
)
from ai.backend.manager.api.gql.domain import Domain
from ai.backend.manager.api.gql.model_deployment.access_token import (
    AccessToken,
    AccessTokenConnection,
    AccessTokenEdge,
    AccessTokenOrderBy,
)
from ai.backend.manager.api.gql.model_deployment.auto_scaling_rule import (
    AutoScalingRule,
)
from ai.backend.manager.api.gql.model_deployment.model_replica import (
    ModelReplicaConnection,
    ReplicaFilter,
    ReplicaOrderBy,
    resolve_replicas,
)
from ai.backend.manager.api.gql.project import Project
from ai.backend.manager.api.gql.types import StrawberryGQLContext
from ai.backend.manager.api.gql.user import User
from ai.backend.manager.data.deployment.creator import NewDeploymentCreator
from ai.backend.manager.data.deployment.modifier import NewDeploymentModifier
from ai.backend.manager.data.deployment.types import (
    DeploymentMetadata,
    DeploymentNetworkSpec,
    DeploymentOrderField,
    ModelDeploymentData,
    ModelDeploymentMetadataInfo,
    ReplicaSpec,
    ReplicaStateData,
)
from ai.backend.manager.errors.user import UserNotFound
from ai.backend.manager.models.gql_models.domain import DomainNode
from ai.backend.manager.models.gql_models.group import GroupNode
from ai.backend.manager.models.gql_models.user import UserNode
from ai.backend.manager.repositories.deployment.types.types import (
    AccessTokenOrderingOptions,
    DeploymentFilterOptions,
    DeploymentOrderingOptions,
    DeploymentStatusFilterType,
)
from ai.backend.manager.repositories.deployment.types.types import (
    DeploymentStatusFilter as RepoDeploymentStatusFilter,
)
from ai.backend.manager.services.deployment.actions.access_token.list_access_tokens import (
    ListAccessTokensAction,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.batch_load_auto_scaling_rules import (
    BatchLoadAutoScalingRulesAction,
)
from ai.backend.manager.services.deployment.actions.batch_load_deployments import (
    BatchLoadDeploymentsAction,
)
from ai.backend.manager.services.deployment.actions.create_deployment import (
    CreateDeploymentAction,
)
from ai.backend.manager.services.deployment.actions.destroy_deployment import (
    DestroyDeploymentAction,
)
from ai.backend.manager.services.deployment.actions.list_deployments import ListDeploymentsAction
from ai.backend.manager.services.deployment.actions.sync_replicas import SyncReplicaAction
from ai.backend.manager.services.deployment.actions.update_deployment import UpdateDeploymentAction
from ai.backend.manager.types import OptionalState, TriState

from .model_revision import (
    CreateModelRevisionInput,
    ModelRevision,
    ModelRevisionConnection,
    ModelRevisionFilter,
    ModelRevisionOrderBy,
    resolve_revisions,
)

DeploymentStatus = strawberry.enum(
    CommonDeploymentStatus,
    name="DeploymentStatus",
    description="Added in 25.16.0. This enum represents the deployment status of a model deployment, indicating its current state.",
)

DeploymentStrategyType = strawberry.enum(
    CommonDeploymentStrategy,
    name="DeploymentStrategyType",
    description="Added in 25.16.0. This enum represents the deployment strategy type of a model deployment, indicating the strategy used for deployment.",
)


@strawberry.type(description="Added in 25.16.0")
class DeploymentStrategy:
    type: DeploymentStrategyType


@strawberry.type(description="Added in 25.16.0")
class ReplicaState:
    _replica_ids: strawberry.Private[list[UUID]]
    desired_replica_count: int

    @strawberry.field
    async def replicas(
        self,
        info: Info[StrawberryGQLContext],
        filter: Optional[ReplicaFilter] = None,
        order_by: Optional[list[ReplicaOrderBy]] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        last: Optional[int] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> ModelReplicaConnection:
        final_filter = ReplicaFilter(ids_in=self._replica_ids)
        if filter:
            final_filter = ReplicaFilter(AND=[final_filter, filter])

        return await resolve_replicas(
            info=info,
            filter=final_filter,
            order_by=order_by,
            before=before,
            after=after,
            first=first,
            last=last,
            limit=limit,
            offset=offset,
        )


@strawberry.type(description="Added in 25.16.0")
class ScalingRule:
    _scaling_rule_ids: strawberry.Private[list[UUID]]

    @strawberry.field
    async def auto_scaling_rules(self, info: Info[StrawberryGQLContext]) -> list[AutoScalingRule]:
        processor = info.context.processors.deployment
        if processor is None:
            raise ModelDeploymentUnavailable(
                "Model Deployment feature is unavailable. Please contact support."
            )

        result = await processor.batch_load_auto_scaling_rules.wait_for_complete(
            BatchLoadAutoScalingRulesAction(auto_scaling_rule_ids=self._scaling_rule_ids)
        )

        return [AutoScalingRule.from_dataclass(rule) for rule in result.data]


@strawberry.type(description="Added in 25.16.0")
class ModelDeploymentMetadata:
    _project_id: strawberry.Private[UUID]
    _domain_name: strawberry.Private[str]
    name: str
    status: DeploymentStatus
    tags: list[str]
    created_at: datetime
    updated_at: datetime

    @strawberry.field
    async def project(self, info: Info[StrawberryGQLContext]) -> Project:
        project_global_id = to_global_id(
            GroupNode, self._project_id, is_target_graphene_object=True
        )
        return Project(id=ID(project_global_id))

    @strawberry.field
    async def domain(self, info: Info[StrawberryGQLContext]) -> Domain:
        domain_global_id = to_global_id(
            DomainNode, self._domain_name, is_target_graphene_object=True
        )
        return Domain(id=ID(domain_global_id))

    @classmethod
    def from_dataclass(cls, data: ModelDeploymentMetadataInfo) -> "ModelDeploymentMetadata":
        return cls(
            name=data.name,
            status=DeploymentStatus(data.status),
            tags=data.tags,
            _project_id=data.project_id,
            _domain_name=data.domain_name,
            created_at=data.created_at,
            updated_at=data.updated_at,
        )


def _convert_gql_revision_ordering_to_repo_ordering(
    order_by: Optional[list[AccessTokenOrderBy]],
) -> AccessTokenOrderingOptions:
    if order_by is None or len(order_by) == 0:
        return AccessTokenOrderingOptions()

    repo_ordering = []
    for order in order_by:
        desc = order.direction == OrderDirection.DESC
        repo_ordering.append((order.field, desc))

    return AccessTokenOrderingOptions(order_by=repo_ordering)


@strawberry.type(description="Added in 25.16.0")
class ModelDeploymentNetworkAccess:
    _access_token_ids: strawberry.Private[Optional[list[UUID]]]
    endpoint_url: Optional[str] = None
    preferred_domain_name: Optional[str] = None
    open_to_public: bool = False

    @strawberry.field
    async def access_tokens(
        self,
        info: Info[StrawberryGQLContext],
        order_by: Optional[list[AccessTokenOrderBy]] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        last: Optional[int] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> AccessTokenConnection:
        """Resolve access tokens using dataloader."""
        repo_ordering = _convert_gql_revision_ordering_to_repo_ordering(order_by)

        pagination_options = build_pagination_options(
            before=before,
            after=after,
            first=first,
            last=last,
            limit=limit,
            offset=offset,
        )

        processor = info.context.processors.deployment
        if processor is None:
            raise ModelDeploymentUnavailable(
                "Model Deployment feature is unavailable. Please contact support."
            )
        action_result = await processor.list_access_tokens.wait_for_complete(
            ListAccessTokensAction(
                pagination=pagination_options,
                ordering=repo_ordering,
            )
        )
        edges = []
        tokens = action_result.data
        total_count = action_result.total_count

        for token in tokens:
            edges.append(
                AccessTokenEdge(
                    node=AccessToken.from_dataclass(token),
                    cursor=to_global_id(AccessToken, token.id),
                )
            )

        page_info = build_page_info(edges, total_count, pagination_options)

        return AccessTokenConnection(
            count=total_count, edges=edges, page_info=page_info.to_strawberry_page_info()
        )

    @classmethod
    def from_dataclass(cls, data: DeploymentNetworkSpec) -> "ModelDeploymentNetworkAccess":
        return cls(
            _access_token_ids=data.access_token_ids,
            endpoint_url=data.url,
            preferred_domain_name=data.preferred_domain_name,
            open_to_public=data.open_to_public,
        )


# Main ModelDeployment Type
@strawberry.type(description="Added in 25.16.0")
class ModelDeployment(Node):
    id: NodeID
    metadata: ModelDeploymentMetadata
    network_access: ModelDeploymentNetworkAccess
    revision: Optional[ModelRevision] = None
    default_deployment_strategy: DeploymentStrategy
    _revision_history_ids: strawberry.Private[list[UUID]]
    _replica_state_data: strawberry.Private[ReplicaStateData]
    _created_user_id: strawberry.Private[UUID]
    _scaling_rule_ids: strawberry.Private[list[UUID]]

    @strawberry.field
    async def created_user(self, info: Info[StrawberryGQLContext]) -> User:
        user_global_id = to_global_id(
            UserNode, self._created_user_id, is_target_graphene_object=True
        )
        return User(id=strawberry.ID(user_global_id))

    @strawberry.field
    async def scaling_rule(self, info: Info[StrawberryGQLContext]) -> ScalingRule:
        return ScalingRule(
            _scaling_rule_ids=self._scaling_rule_ids,
        )

    @strawberry.field
    async def replica_state(self, info: Info[StrawberryGQLContext]) -> ReplicaState:
        return ReplicaState(
            desired_replica_count=self._replica_state_data.desired_replica_count,
            _replica_ids=self._replica_state_data.replica_ids,
        )

    @strawberry.field
    async def revision_history(
        self,
        info: Info[StrawberryGQLContext],
        filter: Optional[ModelRevisionFilter] = None,
        order_by: Optional[list[ModelRevisionOrderBy]] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        first: Optional[int] = None,
        last: Optional[int] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> ModelRevisionConnection:
        final_filter = ModelRevisionFilter(ids_in=self._revision_history_ids)
        if filter:
            final_filter = ModelRevisionFilter(AND=[final_filter, filter])

        return await resolve_revisions(
            info=info,
            filter=final_filter,
            order_by=order_by,
            before=before,
            after=after,
            first=first,
            last=last,
            limit=limit,
            offset=offset,
        )

    @classmethod
    async def batch_load_by_ids(
        cls, ctx: StrawberryGQLContext, deployment_ids: Sequence[UUID]
    ) -> list["ModelDeployment"]:
        """Batch load deployments by their IDs."""
        processor = ctx.processors.deployment
        if processor is None:
            raise ModelDeploymentUnavailable(
                "Model Deployment feature is unavailable. Please contact support."
            )

        result = await processor.batch_load_deployments.wait_for_complete(
            BatchLoadDeploymentsAction(deployment_ids=list(deployment_ids))
        )

        deployment_map = {deployment.id: deployment for deployment in result.data}
        model_deployments = []

        for deployment_id in deployment_ids:
            if deployment_id not in deployment_map:
                raise ModelDeploymentNotFound(f"Deployment with ID {deployment_id} not found")
            model_deployments.append(cls.from_dataclass(deployment_map[deployment_id]))

        return model_deployments

    @classmethod
    def from_dataclass(
        cls,
        data: ModelDeploymentData,
    ) -> "ModelDeployment":
        metadata = ModelDeploymentMetadata(
            name=data.metadata.name,
            status=DeploymentStatus(data.metadata.status),
            tags=data.metadata.tags,
            _project_id=data.metadata.project_id,
            _domain_name=data.metadata.domain_name,
            created_at=data.metadata.created_at,
            updated_at=data.metadata.updated_at,
        )

        return cls(
            id=ID(str(data.id)),
            metadata=metadata,
            network_access=ModelDeploymentNetworkAccess.from_dataclass(data.network_access),
            revision=ModelRevision.from_dataclass(data.revision) if data.revision else None,
            default_deployment_strategy=DeploymentStrategy(
                type=DeploymentStrategyType(data.default_deployment_strategy)
            ),
            _created_user_id=data.created_user_id,
            _revision_history_ids=data.revision_history_ids,
            _scaling_rule_ids=data.scaling_rule_ids,
            _replica_state_data=data.replica_state,
        )


# Filter Types
@strawberry.input(description="Added in 25.16.0")
class DeploymentStatusFilter:
    in_: Optional[list[DeploymentStatus]] = strawberry.field(name="in", default=None)
    equals: Optional[DeploymentStatus] = None


@strawberry.input(description="Added in 25.16.0")
class DeploymentFilter:
    name: Optional[StringFilter] = None
    status: Optional[DeploymentStatusFilter] = None
    open_to_public: Optional[bool] = None
    tags: Optional[StringFilter] = None
    endpoint_url: Optional[StringFilter] = None
    id: Optional[ID] = None

    AND: Optional[list["DeploymentFilter"]] = None
    OR: Optional[list["DeploymentFilter"]] = None
    NOT: Optional[list["DeploymentFilter"]] = None

    def to_repo_filter(self) -> DeploymentFilterOptions:
        repo_filter = DeploymentFilterOptions()

        repo_filter.name = self.name
        repo_filter.open_to_public = self.open_to_public
        repo_filter.tags = self.tags
        repo_filter.endpoint_url = self.endpoint_url
        repo_filter.id = UUID(self.id) if self.id else None
        if self.status:
            if self.status.in_ is not None:
                repo_filter.status = RepoDeploymentStatusFilter(
                    type=DeploymentStatusFilterType.IN,
                    values=[CommonDeploymentStatus(status) for status in self.status.in_],
                )
            elif self.status.equals is not None:
                repo_filter.status = RepoDeploymentStatusFilter(
                    type=DeploymentStatusFilterType.EQUALS,
                    values=[CommonDeploymentStatus(self.status.equals)],
                )

        # Handle logical operations
        if self.AND:
            repo_filter.AND = [f.to_repo_filter() for f in self.AND]
        if self.OR:
            repo_filter.OR = [f.to_repo_filter() for f in self.OR]
        if self.NOT:
            repo_filter.NOT = [f.to_repo_filter() for f in self.NOT]

        return repo_filter


@strawberry.input(description="Added in 25.16.0")
class DeploymentOrderBy:
    field: DeploymentOrderField
    direction: OrderDirection = OrderDirection.DESC


# Payload Types
@strawberry.type(description="Added in 25.16.0")
class CreateModelDeploymentPayload:
    deployment: ModelDeployment


@strawberry.type(description="Added in 25.16.0")
class UpdateModelDeploymentPayload:
    deployment: ModelDeployment


@strawberry.type(description="Added in 25.16.0")
class DeleteModelDeploymentPayload:
    id: ID


@strawberry.type(description="Added in 25.16.0")
class DeploymentStatusChangedPayload:
    deployment: ModelDeployment


# Input Types
@strawberry.input(description="Added in 25.16.0")
class ModelDeploymentMetadataInput:
    project_id: ID
    domain_name: str
    name: Optional[str] = None
    tags: Optional[list[str]] = None


@strawberry.input(description="Added in 25.16.0")
class ModelDeploymentNetworkAccessInput:
    preferred_domain_name: Optional[str] = None
    open_to_public: bool = False

    def to_network_spec(self) -> DeploymentNetworkSpec:
        return DeploymentNetworkSpec(
            open_to_public=self.open_to_public,
            preferred_domain_name=self.preferred_domain_name,
        )


@strawberry.input(description="Added in 25.16.0")
class DeploymentStrategyInput:
    type: DeploymentStrategyType


@strawberry.input(description="Added in 25.16.0")
class CreateModelDeploymentInput:
    metadata: ModelDeploymentMetadataInput
    network_access: ModelDeploymentNetworkAccessInput
    default_deployment_strategy: DeploymentStrategyInput
    desired_replica_count: int
    initial_revision: CreateModelRevisionInput

    def to_creator(self) -> NewDeploymentCreator:
        name = self.metadata.name or f"deployment-{uuid4().hex[:8]}"
        tag = ",".join(self.metadata.tags) if self.metadata.tags else None
        user_data = current_user()
        if user_data is None:
            raise UserNotFound("User not found in context")
        metadata_for_creator = DeploymentMetadata(
            name=name,
            domain=self.metadata.domain_name,
            project=UUID(str(self.metadata.project_id)),
            resource_group=self.initial_revision.resource_config.resource_group.name,
            created_user=user_data.user_id,
            session_owner=user_data.user_id,
            created_at=None,
            tag=tag,
        )
        return NewDeploymentCreator(
            metadata=metadata_for_creator,
            replica_spec=ReplicaSpec(replica_count=self.desired_replica_count),
            network=self.network_access.to_network_spec(),
            model_revision=self.initial_revision.to_model_revision_creator(),
        )


@strawberry.input(description="Added in 25.16.0")
class UpdateModelDeploymentInput:
    id: ID
    open_to_public: Optional[bool] = None
    tags: Optional[list[str]] = None
    default_deployment_strategy: Optional[DeploymentStrategyInput] = None
    active_revision_id: Optional[ID] = None
    desired_replica_count: Optional[int] = None
    name: Optional[str] = None
    preferred_domain_name: Optional[str] = None

    def to_modifier(self) -> NewDeploymentModifier:
        strategy_type = None
        if self.default_deployment_strategy is not None:
            strategy_type = CommonDeploymentStrategy(self.default_deployment_strategy.type)
        return NewDeploymentModifier(
            open_to_public=OptionalState[bool].from_graphql(self.open_to_public),
            tags=OptionalState[list[str]].from_graphql(self.tags),
            default_deployment_strategy=OptionalState[CommonDeploymentStrategy].from_graphql(
                strategy_type
            ),
            active_revision_id=OptionalState[UUID].from_graphql(UUID(self.active_revision_id)),
            desired_replica_count=OptionalState[int].from_graphql(self.desired_replica_count),
            name=OptionalState[str].from_graphql(self.name),
            preferred_domain_name=TriState[str].from_graphql(self.preferred_domain_name),
        )


@strawberry.input(description="Added in 25.16.0")
class DeleteModelDeploymentInput:
    id: ID


ModelDeploymentEdge = Edge[ModelDeployment]


# Connection types for Relay support
@strawberry.type(description="Added in 25.16.0")
class ModelDeploymentConnection(Connection[ModelDeployment]):
    count: int

    def __init__(self, *args, count: int, **kwargs):
        super().__init__(*args, **kwargs)
        self.count = count


def _convert_gql_deployment_ordering_to_repo(
    order_by: Optional[list[DeploymentOrderBy]],
) -> DeploymentOrderingOptions:
    if order_by is None or len(order_by) == 0:
        return DeploymentOrderingOptions()

    repo_ordering = []
    for order in order_by:
        desc = order.direction == OrderDirection.DESC
        repo_ordering.append((order.field, desc))
    return DeploymentOrderingOptions(order_by=repo_ordering)


async def resolve_deployments(
    info: Info[StrawberryGQLContext],
    filter: Optional[DeploymentFilter] = None,
    order_by: Optional[list[DeploymentOrderBy]] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> ModelDeploymentConnection:
    repo_filter = None
    if filter:
        repo_filter = filter.to_repo_filter()

    repo_ordering = _convert_gql_deployment_ordering_to_repo(order_by)

    pagination_options = build_pagination_options(
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )

    processor = info.context.processors.deployment
    if processor is None:
        raise ModelDeploymentUnavailable(
            "Model Deployment feature is unavailable. Please contact support."
        )
    action_result = await processor.list_deployments.wait_for_complete(
        ListDeploymentsAction(
            pagination=pagination_options, ordering=repo_ordering, filters=repo_filter
        )
    )
    edges = []
    for deployment in action_result.data:
        edges.append(
            ModelDeploymentEdge(
                node=ModelDeployment.from_dataclass(deployment), cursor=str(deployment.id)
            )
        )
    page_info = build_page_info(
        edges=edges,
        total_count=action_result.total_count,
        pagination_options=pagination_options,
    )

    connection = ModelDeploymentConnection(
        count=action_result.total_count,
        edges=edges,
        page_info=page_info.to_strawberry_page_info(),
    )
    return connection


# Resolvers
@strawberry.field(description="Added in 25.16.0")
async def deployments(
    info: Info[StrawberryGQLContext],
    filter: Optional[DeploymentFilter] = None,
    order_by: Optional[list[DeploymentOrderBy]] = None,
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> ModelDeploymentConnection:
    """List deployments with optional filtering and pagination."""

    return await resolve_deployments(
        info=info,
        filter=filter,
        order_by=order_by,
        before=before,
        after=after,
        first=first,
        last=last,
        limit=limit,
        offset=offset,
    )


@strawberry.field(description="Added in 25.16.0")
async def deployment(id: ID, info: Info[StrawberryGQLContext]) -> Optional[ModelDeployment]:
    """Get a specific deployment by ID."""
    _, deployment_id = resolve_global_id(id)
    deployment_dataloader = info.context.dataloader_registry.get_loader(
        ModelDeployment.batch_load_by_ids, info.context
    )
    deployment: list[ModelDeployment] = await deployment_dataloader.load(deployment_id)

    return deployment[0]


@strawberry.mutation(description="Added in 25.16.0")
async def create_model_deployment(
    input: CreateModelDeploymentInput, info: Info[StrawberryGQLContext]
) -> "CreateModelDeploymentPayload":
    """Create a new model deployment."""

    processor = info.context.processors.deployment
    if processor is None:
        raise ModelDeploymentUnavailable(
            "Model Deployment feature is unavailable. Please contact support."
        )

    result = await processor.create_deployment.wait_for_complete(
        CreateDeploymentAction(creator=input.to_creator())
    )

    return CreateModelDeploymentPayload(deployment=ModelDeployment.from_dataclass(result.data))


@strawberry.mutation(description="Added in 25.16.0")
async def update_model_deployment(
    input: UpdateModelDeploymentInput, info: Info[StrawberryGQLContext]
) -> UpdateModelDeploymentPayload:
    """Update an existing model deployment."""
    _, deployment_id = resolve_global_id(input.id)
    deployment_processor = info.context.processors.deployment
    if deployment_processor is None:
        raise ModelDeploymentUnavailable(
            "Model Deployment feature is unavailable. Please contact support."
        )
    action_result = await deployment_processor.update_deployment.wait_for_complete(
        UpdateDeploymentAction(deployment_id=UUID(deployment_id), modifier=input.to_modifier())
    )
    return UpdateModelDeploymentPayload(
        deployment=ModelDeployment.from_dataclass(action_result.data)
    )


@strawberry.mutation(description="Added in 25.16.0")
async def delete_model_deployment(
    input: DeleteModelDeploymentInput, info: Info[StrawberryGQLContext]
) -> DeleteModelDeploymentPayload:
    """Delete a model deployment."""
    _, deployment_id = resolve_global_id(input.id)
    deployment_processor = info.context.processors.deployment
    if deployment_processor is None:
        raise ModelDeploymentUnavailable(
            "Model Deployment feature is unavailable. Please contact support."
        )
    _ = await deployment_processor.destroy_deployment.wait_for_complete(
        DestroyDeploymentAction(endpoint_id=UUID(deployment_id))
    )
    return DeleteModelDeploymentPayload(id=input.id)


@strawberry.subscription(description="Added in 25.16.0")
async def deployment_status_changed(
    deployment_id: ID, info: Info[StrawberryGQLContext]
) -> AsyncGenerator[DeploymentStatusChangedPayload, None]:
    """Subscribe to deployment status changes."""
    # Mock implementation
    # In real implementation, this would yield artifacts when status changes
    if False:  # Placeholder to make this a generator
        yield DeploymentStatusChangedPayload(deployment_id=deployment_id)


@strawberry.input(description="Added in 25.16.0")
class SyncReplicaInput:
    model_deployment_id: ID


@strawberry.type(description="Added in 25.16.0")
class SyncReplicaPayload:
    success: bool


@strawberry.mutation(
    description="Added in 25.16.0. Force syncs up-to-date replica information. In normal situations this will be automatically handled by Backend.AI schedulers"
)
async def sync_replicas(
    input: SyncReplicaInput, info: Info[StrawberryGQLContext]
) -> SyncReplicaPayload:
    _, deployment_id = resolve_global_id(input.model_deployment_id)
    deployment_processor = info.context.processors.deployment
    if deployment_processor is None:
        raise ModelDeploymentUnavailable(
            "Model Deployment feature is unavailable. Please contact support."
        )
    await deployment_processor.sync_replicas.wait_for_complete(
        SyncReplicaAction(deployment_id=UUID(deployment_id))
    )
    return SyncReplicaPayload(success=True)



from typing import Final

DEFAULT_CHUNK_SIZE: Final = 256 * 1024  # 256 KiB
DEFAULT_INFLIGHT_CHUNKS: Final = 8


MANAGER_LOCAL_CFG_OVERRIDE_ENVS: list[tuple[tuple[str, ...], str]] = [
    (("etcd", "namespace"), "BACKEND_NAMESPACE"),
    (("etcd", "addr"), "BACKEND_ETCD_ADDR"),
    (("etcd", "user"), "BACKEND_ETCD_USER"),
    (("etcd", "password"), "BACKEND_ETCD_PASSWORD"),
    (("db", "addr"), "BACKEND_DB_ADDR"),
    (("db", "name"), "BACKEND_DB_NAME"),
    (("db", "user"), "BACKEND_DB_USER"),
    (("db", "password"), "BACKEND_DB_PASSWORD"),
    (("manager", "num-proc"), "BACKEND_MANAGER_NPROC"),
    (("manager", "ssl-cert"), "BACKEND_SSL_CERT"),
    (("manager", "ssl-privkey"), "BACKEND_SSL_KEY"),
    (("manager", "pid-file"), "BACKEND_PID_FILE"),
    (("manager", "api-listen-addr", "host"), "BACKEND_SERVICE_IP"),
    (("manager", "api-listen-addr", "port"), "BACKEND_SERVICE_PORT"),
    (("manager", "event-listen-addr", "host"), "BACKEND_ADVERTISED_MANAGER_HOST"),
    (("manager", "event-listen-addr", "port"), "BACKEND_EVENTS_PORT"),
    (("docker-registry", "ssl-verify"), "BACKEND_SKIP_SSLCERT_VALIDATION"),
]
